# @omlish-lite
# ruff: noqa: UP006 UP007 UP043 UP045
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
"""
https://github.com/python/cpython/blob/9b335cc8104dd83a5a1343dc649d1f3606682098/Lib/http/client.py
"""
import collections.abc
import email.parser
import enum
import http
import io
import typing as ta
import urllib.parse

from ....lite.check import check
from ..io import CoroHttpIo
from .errors import CoroHttpClientErrors
from .headers import CoroHttpClientHeaders
from .response import CoroHttpClientResponse
from .status import CoroHttpClientStatusLine
from .validation import CoroHttpClientValidation


##


class CoroHttpClientConnection:
    """
    HTTPConnection goes through a number of "states", which define when a client may legally make another request or
    fetch the response for a particular request. This diagram details these state transitions:

        (null)
          |
          | HTTPConnection()
          v
        Idle
          |
          | put_request()
          v
        Request-started
          |
          | ( put_header() )*  end_headers()
          v
        Request-sent
          |______________________________
          |                              | get_response() raises
          | response = get_response()     | ConnectionError
          v                              v
        Unread-response                Idle
        [Response-headers-read]
          |_____________________
          |                     |
          | response.read()     | put_request()
          v                     v
        Idle                  Req-started-unread-response
                         ______/|
                       /        |
       response.read() |        | ( put_header() )*  end_headers()
                       v        v
           Request-started    Req-sent-unread-response
                                |
                                | response.read()
                                v
                              Request-sent

    This diagram presents the following rules:
      -- a second request may not be started until {response-headers-read}
      -- a response [object] cannot be retrieved until {request-sent}
      -- there is no differentiation between an unread response body and a
         partially read response body

    Logical State                _state       _response
    -------------                -------      ----------
    Idle                         IDLE         None
    Request-started              REQ_STARTED  None
    Request-sent                 REQ_SENT     None
    Unread-response              IDLE         <response_class>
    Req-started-unread-response  REQ_STARTED  <response_class>
    Req-sent-unread-response     REQ_SENT     <response_class>
    """

    _http_version = 11
    _http_version_str = 'HTTP/1.1'

    HTTP_PORT: ta.ClassVar[int] = 80
    HTTPS_PORT: ta.ClassVar[int] = 443

    DEFAULT_PORT: ta.ClassVar[int] = HTTP_PORT

    class _NOT_SET:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise NotImplementedError

    class _State(enum.Enum):
        IDLE = 'Idle'
        REQ_STARTED = 'Request-started'
        REQ_SENT = 'Request-sent'

    def __init__(
            self,
            host: str,
            port: ta.Optional[int] = None,
            *,
            timeout: ta.Union[float, ta.Type[_NOT_SET], None] = _NOT_SET,
            source_address: ta.Optional[str] = None,
            block_size: int = 8192,
            auto_open: bool = True,
            default_port: ta.Optional[int] = None,
    ) -> None:
        super().__init__()

        self._timeout = timeout
        self._source_address = source_address
        self._block_size = block_size
        self._auto_open = auto_open
        if default_port is None:
            default_port = self.DEFAULT_PORT
        self._default_port = default_port

        self._connected = False
        self._buffer: ta.List[bytes] = []
        self._response: ta.Optional[CoroHttpClientResponse] = None
        self._state = self._State.IDLE
        self._method: ta.Optional[str] = None

        self._tunnel_host: ta.Optional[str] = None
        self._tunnel_port: ta.Optional[int] = None
        self._tunnel_headers: ta.Dict[str, str] = {}
        self._raw_proxy_headers: ta.Optional[ta.Sequence[bytes]] = None

        (self._host, self._port) = self._get_hostport(host, port)

        CoroHttpClientValidation.validate_host(self._host)

    @property
    def http_version(self) -> int:
        return self._http_version

    #

    def _get_hostport(self, host: str, port: ta.Optional[int]) -> ta.Tuple[str, int]:
        if port is None:
            i = host.rfind(':')
            j = host.rfind(']')  # ipv6 addresses have [...]
            if i > j:
                try:
                    port = int(host[i + 1:])
                except ValueError:
                    if host[i + 1:] == '':  # http://foo.com:/ == http://foo.com/
                        port = self._default_port
                    else:
                        raise CoroHttpClientErrors.InvalidUrlError(f"non-numeric port: '{host[i + 1:]}'") from None
                host = host[:i]
            else:
                port = self._default_port

        if host and host[0] == '[' and host[-1] == ']':
            host = host[1:-1]

        return (host, port)

    def _wrap_ipv6(self, ip: bytes) -> bytes:
        if b':' in ip and ip[0] != b'['[0]:
            return b'[' + ip + b']'
        return ip

    #

    def set_tunnel(
            self,
            host: str,
            port: ta.Optional[int] = None,
            headers: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> None:
        """
        Set up host and port for HTTP CONNECT tunnelling.

        In a connection that uses HTTP CONNECT tunnelling, the host passed to the constructor is used as a proxy server
        that relays all communication to the endpoint passed to `set_tunnel`. This done by sending an HTTP CONNECT
        request to the proxy server when the connection is established.

        This method must be called before the HTTP connection has been established.

        The headers argument should be a mapping of extra HTTP headers to send with the CONNECT request.

        As HTTP/1.1 is used for HTTP CONNECT tunnelling request, as per the RFC
        (https://tools.ietf.org/html/rfc7231#section-4.3.6), a HTTP Host: header must be provided, matching the
        authority-form of the request target provided as the destination for the CONNECT request. If a HTTP Host: header
        is not provided via the headers argument, one is generated and transmitted automatically.
        """

        if self._connected:
            raise RuntimeError("Can't set up tunnel for established connection")

        self._tunnel_host, self._tunnel_port = self._get_hostport(host, port)

        if headers:
            self._tunnel_headers = dict(headers)
        else:
            self._tunnel_headers.clear()

        if not any(header.lower() == 'host' for header in self._tunnel_headers):
            encoded_host = self._tunnel_host.encode('idna').decode('ascii')
            self._tunnel_headers['Host'] = f'{encoded_host}:{self._tunnel_port:d}'

    def _tunnel(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        connect = b'CONNECT %s:%d %s\r\n' % (
            self._wrap_ipv6(check.not_none(self._tunnel_host).encode('idna')),
            check.not_none(self._tunnel_port),
            self._http_version_str.encode('ascii'),
        )

        headers = [connect]
        for header, value in self._tunnel_headers.items():
            headers.append(f'{header}: {value}\r\n'.encode('latin-1'))
        headers.append(b'\r\n')

        # Making a single send() call instead of one per line encourages the host OS to use a more optimal packet size
        # instead of potentially emitting a series of small packets.
        yield from self.send(b''.join(headers))
        del headers

        try:
            (version, code, message) = (yield from CoroHttpClientStatusLine.read())
        except CoroHttpClientErrors.BadStatusLineError:  # noqa
            # self._close_conn()
            raise

        self._raw_proxy_headers = yield from CoroHttpClientHeaders.read_headers()

        if code != http.HTTPStatus.OK:
            yield from self.close()
            raise OSError(f'Tunnel connection failed: {code} {message.strip()}')

    def get_proxy_response_headers(self) -> ta.Optional[email.message.Message]:
        """
        Returns a dictionary with the headers of the response received from the proxy server to the CONNECT request sent
        to set the tunnel.

        If the CONNECT request was not sent, the method returns None.
        """

        return (
            CoroHttpClientHeaders.parse_header_lines(self._raw_proxy_headers)
            if self._raw_proxy_headers is not None
            else None
        )

    #

    def connect(self) -> ta.Generator[CoroHttpIo.Io, None, None]:
        """Connect to the host and port specified in __init__."""

        if self._connected:
            return

        check.none((yield CoroHttpIo.ConnectIo(
            ((self._host, self._port),),
            dict(
                source_address=self._source_address,
                **(dict(timeout=self._timeout) if self._timeout is not self._NOT_SET else {}),
            ),
            server_hostname=self._tunnel_host if self._tunnel_host else self._host,
        )))

        self._connected = True

        if self._tunnel_host:
            yield from self._tunnel()

    #

    def close(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        """Close the connection to the HTTP server."""

        self._state = self._State.IDLE

        try:
            if self._connected:
                yield CoroHttpIo.CloseIo()  # Close it manually... there may be other refs
                self._connected = False

        finally:
            response = self._response
            if response:
                self._response = None
                response.close()

    #

    @staticmethod
    def _is_text_io(stream: ta.Any) -> bool:
        """Test whether a file-like object is a text or a binary stream."""

        return isinstance(stream, io.TextIOBase)

    def send(self, data: ta.Any) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        """
        Send 'data' to the server. ``data`` can be a string object, a bytes object, an array object, a file-like object
        that supports a .read() method, or an iterable object.
        """

        if not self._connected:
            if self._auto_open:
                yield from self.connect()
            else:
                raise CoroHttpClientErrors.NotConnectedError

        check.state(self._connected)

        if hasattr(data, 'read'):
            encode = self._is_text_io(data)
            while data_block := data.read(self._block_size):
                if encode:
                    data_block = data_block.encode('iso-8859-1')
                check.none((yield CoroHttpIo.WriteIo(data_block)))
            return

        if isinstance(data, (bytes, bytearray)):
            check.none((yield CoroHttpIo.WriteIo(data)))

        elif isinstance(data, collections.abc.Iterable):
            for d in data:
                check.none((yield CoroHttpIo.WriteIo(d)))

        else:
            raise TypeError(f'data should be a bytes-like object or an iterable, got {type(data)!r}') from None

    def _output(self, s: bytes) -> None:
        """
        Add a line of output to the current request buffer.

        Assumes that the line does *not* end with \\r\\n.
        """

        self._buffer.append(s)

    def _read_readable(self, readable: ta.Union[ta.IO, ta.TextIO]) -> ta.Iterator[bytes]:
        while data := readable.read(self._block_size):
            if isinstance(data, str):
                yield data.encode('iso-8859-1')
            else:
                yield data

    def _send_output(
            self,
            message_body: ta.Optional[ta.Any] = None,
            encode_chunked: bool = False,
    ) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        """
        Send the currently buffered request and clear the buffer.

        Appends an extra \\r\\n to the buffer. A message_body may be specified, to be appended to the request.
        """

        self._buffer.extend((b'', b''))
        msg = b'\r\n'.join(self._buffer)
        del self._buffer[:]
        yield from self.send(msg)

        chunks: ta.Iterable[bytes]
        if message_body is not None:
            # Create a consistent interface to message_body
            if hasattr(message_body, 'read'):
                # Let file-like take precedence over byte-like. This is needed to allow the current position of mmap'ed
                # files to be taken into account.
                chunks = self._read_readable(message_body)

            else:
                try:
                    # This is solely to check to see if message_body implements the buffer API. it /would/ be easier to
                    # capture if PyObject_CheckBuffer was exposed to Python.
                    memoryview(message_body)

                except TypeError:
                    try:
                        chunks = iter(message_body)
                    except TypeError as e:
                        raise TypeError(
                            f'message_body should be a bytes-like object or an iterable, got {type(message_body)!r}',
                        ) from e

                else:
                    # The object implements the buffer interface and can be passed directly into socket methods
                    chunks = (message_body,)

            for chunk in chunks:
                if not chunk:
                    continue

                if encode_chunked and self._http_version == 11:
                    # Chunked encoding
                    chunk = f'{len(chunk):X}\r\n'.encode('ascii') + chunk + b'\r\n'
                yield from self.send(chunk)

            if encode_chunked and self._http_version == 11:
                # End chunked transfer
                yield from self.send(b'0\r\n\r\n')

    #

    @staticmethod
    def _strip_ipv6_iface(enc_name: bytes) -> bytes:
        """Remove interface scope from IPv6 address."""

        enc_name, percent, _ = enc_name.partition(b'%')
        if percent:
            check.state(enc_name.startswith(b'['))
            enc_name += b']'
        return enc_name

    def put_request(
            self,
            method: str,
            url: str,
            *,
            skip_host: bool = False,
            skip_accept_encoding: bool = False,
    ) -> None:
        """
        Send a request to the server.

        'method' specifies an HTTP request method, e.g. 'GET'.
        'url' specifies the object being requested, e.g. '/index.html'.
        'skip_host' if True does not add automatically a 'Host:' header
        'skip_accept_encoding' if True does not add automatically an 'Accept-Encoding:' header
        """

        # If a prior response has been completed, then forget about it.
        if self._response and self._response.is_closed():
            self._response = None

        # In certain cases, we cannot issue another request on this connection.
        # this occurs when:
        #   1) we are in the process of sending a request. (_CS_REQ_STARTED)
        #   2) a response to a previous request has signalled that it is going to close the connection upon completion.
        #   3) the headers for the previous response have not been read, thus we cannot determine whether point (2) is
        #      true. (_CS_REQ_SENT)
        #
        # If there is no prior response, then we can request at will.
        #
        # If point (2) is true, then we will have passed the socket to the response (effectively meaning, "there is no
        # prior response"), and will open a new one when a new request is made.
        #
        # Note: if a prior response exists, then we *can* start a new request. We are not allowed to begin fetching the
        #       response to this new request, however, until that prior response is complete.
        #
        if self._state == self._State.IDLE:
            self._state = self._State.REQ_STARTED
        else:
            raise CoroHttpClientErrors.CannotSendRequestError(self._state)

        CoroHttpClientValidation.validate_method(method)

        # Save the method for use later in the response phase
        self._method = method

        url = url or '/'
        CoroHttpClientValidation.validate_path(url)

        request = f'{method} {url} {self._http_version_str}'

        self._output(self._encode_request(request))

        if self._http_version == 11:
            # Issue some standard headers for better HTTP/1.1 compliance

            if not skip_host:
                # This header is issued *only* for HTTP/1.1 connections. more specifically, this means it is only issued
                # when the client uses the new HTTPConnection() class. backwards-compat clients will be using HTTP/1.0
                # and those clients may be issuing this header themselves. we should NOT issue it twice; some web
                # servers (such as Apache) barf when they see two Host: headers

                # If we need a non-standard port,include it in the header. If the request is going through a proxy, but
                # the host of the actual URL, not the host of the proxy.
                netloc = ''
                if url.startswith('http'):
                    netloc = urllib.parse.urlsplit(url).netloc

                if netloc:
                    try:
                        netloc_enc = netloc.encode('ascii')
                    except UnicodeEncodeError:
                        netloc_enc = netloc.encode('idna')
                    self.put_header('Host', self._strip_ipv6_iface(netloc_enc))
                else:
                    if self._tunnel_host:
                        host = self._tunnel_host
                        port = self._tunnel_port
                    else:
                        host = self._host
                        port = self._port

                    try:
                        host_enc = host.encode('ascii')
                    except UnicodeEncodeError:
                        host_enc = host.encode('idna')

                    # As per RFC 273, IPv6 address should be wrapped with [] when used as Host header
                    host_enc = self._wrap_ipv6(host_enc)
                    if ':' in host:
                        host_enc = self._strip_ipv6_iface(host_enc)

                    if port == self._default_port:
                        self.put_header('Host', host_enc)
                    else:
                        self.put_header('Host', f"{host_enc.decode('ascii')}:{port}")

            # NOTE: We are assuming that clients will not attempt to set these headers since *this* library must deal
            # with the consequences. this also means that when the supporting libraries are updated to recognize other
            # forms, then this code should be changed (removed or updated).

            # We only want a Content-Encoding of "identity" since we don't support encodings such as x-gzip or
            # x-deflate.
            if not skip_accept_encoding:
                self.put_header('Accept-Encoding', 'identity')

            # We can accept "chunked" Transfer-Encodings, but no others.
            # NOTE: no TE header implies *only* "chunked"
            #self.put_header('TE', 'chunked')

            # If TE is supplied in the header, then it must appear in a Connection header.
            #self.put_header('Connection', 'TE')

        else:
            # For HTTP/1.0, the server will assume "not chunked"
            pass

    def _encode_request(self, request: str) -> bytes:
        # ASCII also helps prevent CVE-2019-9740.
        return request.encode('ascii')

    #

    def put_header(self, header: ta.Union[str, bytes], *values: ta.Union[bytes, str, int]) -> None:
        """
        Send a request header line to the server.

        For example: h.put_header('Accept', 'text/html')
        """

        if self._state != self._State.REQ_STARTED:
            raise CoroHttpClientErrors.CannotSendHeaderError

        if hasattr(header, 'encode'):
            bh = header.encode('ascii')
        else:
            bh = header

        CoroHttpClientValidation.validate_header_name(bh)

        bvs = []
        for one_value in values:
            if hasattr(one_value, 'encode'):
                bv = one_value.encode('latin-1')
            elif isinstance(one_value, int):
                bv = str(one_value).encode('ascii')
            else:
                bv = one_value

            CoroHttpClientValidation.validate_header_value(bv)
            bvs.append(bv)

        value = b'\r\n\t'.join(bvs)
        bh = bh + b': ' + value
        self._output(bh)

    def end_headers(
            self,
            message_body: ta.Optional[ta.Any] = None,
            *,
            encode_chunked: bool = False,
    ) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        """
        Indicate that the last header line has been sent to the server.

        This method sends the request to the server. The optional message_body argument can be used to pass a message
        body associated with the request.
        """

        if self._state == self._State.REQ_STARTED:
            self._state = self._State.REQ_SENT
        else:
            raise CoroHttpClientErrors.CannotSendHeaderError

        yield from self._send_output(message_body, encode_chunked=encode_chunked)

    #

    def request(
            self,
            method: str,
            url: str,
            body: ta.Optional[ta.Any] = None,
            headers: ta.Optional[ta.Mapping[str, str]] = None,
            *,
            encode_chunked: bool = False,
    ) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        """Send a complete request to the server."""

        yield from self._send_request(method, url, body, dict(headers or {}), encode_chunked)

    _METHODS_EXPECTING_BODY: ta.ClassVar[ta.Container[str]] = {'PATCH', 'POST', 'PUT'}

    @classmethod
    def _get_content_length(
            cls,
            body: ta.Optional[ta.Any],
            method: str,
    ) -> ta.Optional[int]:
        """
        Get the content-length based on the body.

        If the body is None, we set Content-Length: 0 for methods that expect a body (RFC 7230, Section 3.3.2). We also
        set the Content-Length for any method if the body is a str or bytes-like object and not a file.
        """

        if body is None:
            # Do an explicit check for not None here to distinguish between unset and set but empty
            if method.upper() in cls._METHODS_EXPECTING_BODY:
                return 0
            else:
                return None

        if hasattr(body, 'read'):
            # File-like object.
            return None

        try:
            # Does it implement the buffer protocol (bytes, bytearray, array)?
            mv = memoryview(body)
            return mv.nbytes
        except TypeError:
            pass

        if isinstance(body, str):
            return len(body)

        return None

    @staticmethod
    def _encode(data: str, name: str = 'data') -> bytes:
        """Call data.encode("latin-1") but show a better error message."""

        try:
            return data.encode('latin-1')

        except UnicodeEncodeError as err:
            raise UnicodeEncodeError(
                err.encoding,
                err.object,
                err.start,
                err.end,
                "%s (%.20r) is not valid Latin-1. Use %s.encode('utf-8') if you want to send it encoded in UTF-8." % (  # noqa
                    name.title(),
                    data[err.start:err.end],
                    name,
                ),
            ) from None

    def _send_request(
            self,
            method: str,
            url: str,
            body: ta.Optional[ta.Any],
            headers: ta.Mapping[str, str],
            encode_chunked: bool,
    ) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        # Honor explicitly requested Host: and Accept-Encoding: headers.
        header_names = frozenset(k.lower() for k in headers)
        skips = {}
        if 'host' in header_names:
            skips['skip_host'] = True
        if 'accept-encoding' in header_names:
            skips['skip_accept_encoding'] = True

        self.put_request(method, url, **skips)

        # Chunked encoding will happen if HTTP/1.1 is used and either the caller passes encode_chunked=True or the
        # following conditions hold:
        #  1) Content-Length has not been explicitly set
        #  2) The body is a file or iterable, but not a str or bytes-like
        #  3) Transfer-Encoding has NOT been explicitly set by the caller

        if 'content-length' not in header_names:
            # Only chunk body if not explicitly set for backwards compatibility, assuming the client code is already
            # handling the chunking
            if 'transfer-encoding' not in header_names:
                # If Content-Length cannot be automatically determined, fall back to chunked encoding
                encode_chunked = False
                content_length = self._get_content_length(body, method)
                if content_length is None:
                    if body is not None:
                        encode_chunked = True
                        self.put_header('Transfer-Encoding', 'chunked')
                else:
                    self.put_header('Content-Length', str(content_length))
        else:
            encode_chunked = False

        for hdr, value in headers.items():
            self.put_header(hdr, value)

        if isinstance(body, str):
            # RFC 2616 Section 3.7.1 says that text default has a default charset of iso-8859-1.
            body = self._encode(body, 'body')

        yield from self.end_headers(body, encode_chunked=encode_chunked)

    #

    def _new_response(self) -> CoroHttpClientResponse:
        return CoroHttpClientResponse(check.not_none(self._method))

    def get_response(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], CoroHttpClientResponse]:
        """
        Get the response from the server.

        If the HTTPConnection is in the correct state, returns an instance of HttpResponse or of whatever object is
        returned by the response_class variable.

        If a request has not been sent or if a previous response has not be handled, ResponseNotReady is raised. If the
        HTTP response indicates that the connection should be closed, then it will be closed before the response is
        returned. When the connection is closed, the underlying socket is closed.
        """

        # If a prior response has been completed, then forget about it.
        if self._response and self._response.is_closed():
            self._response = None

        # If a prior response exists, then it must be completed (otherwise, we cannot read this response's header to
        # determine the connection-close behavior).
        #
        # NOTE: If a prior response existed, but was connection-close, then the socket and response were made
        # independent of this HTTPConnection object since a new request requires that we open a whole new connection.
        #
        # This means the prior response had one of two states:
        #  1) will_close: this connection was reset and the prior socket and response operate independently
        #  2) persistent: the response was retained and we await its is_closed() status to become true.
        if self._state != self._State.REQ_SENT or self._response:
            raise CoroHttpClientErrors.ResponseNotReadyError(self._state)

        resp = self._new_response()
        resp_state = resp._state  # noqa

        try:
            try:
                yield from resp._begin()  # noqa
            except ConnectionError:
                yield from self.close()
                raise

            check.state(hasattr(resp_state, 'will_close'))
            self._state = self._State.IDLE

            if resp_state.will_close:
                # This effectively passes the connection to the response
                yield from self.close()
            else:
                # Remember this, so we can tell when it is complete
                self._response = resp

            return resp

        except:  # noqa
            resp.close()
            raise



import decimal
import fractions
import typing as ta

from ... import check
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.typemap import TypeMapMarshalerFactory
from ..factories.typemap import TypeMapUnmarshalerFactory


##


class ComplexMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: complex) -> Value:
        return [o.real, o.imag]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        real, imag = check.isinstance(v, list)
        return complex(real, imag)


class DecimalMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: decimal.Decimal) -> Value:
        return str(o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return decimal.Decimal(check.isinstance(v, str))


class FractionMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: fractions.Fraction) -> Value:
        return [o.numerator, o.denominator]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        num, denom = check.isinstance(v, list)
        return fractions.Fraction(num, denom)


NUMBERS_MARSHALER_FACTORY = TypeMapMarshalerFactory({
    complex: ComplexMarshalerUnmarshaler(),
    decimal.Decimal: DecimalMarshalerUnmarshaler(),
    fractions.Fraction: FractionMarshalerUnmarshaler(),
})

NUMBERS_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({
    complex: ComplexMarshalerUnmarshaler(),
    decimal.Decimal: DecimalMarshalerUnmarshaler(),
    fractions.Fraction: FractionMarshalerUnmarshaler(),
})



import typing as ta

from ..codecs import make_object_lazy_loaded_codec
from ..codecs import make_str_object_codec
from .parsing import parse


##


def dumps(obj: ta.Any) -> str:
    # return json.dumps(obj)
    raise NotImplementedError


def loads(s: str) -> ta.Any:
    return parse(s)


##


EDN_CODEC = make_str_object_codec('edn', dumps, loads)

# @omlish-manifest
_EDN_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'EDN_CODEC', EDN_CODEC)



import typing as ta

from ..providers import AsyncFnProvider
from ..providers import ConstProvider
from ..providers import CtorProvider
from ..providers import FnProvider
from ..providers import LinkProvider
from ..providers import Provider
from ..scopes import ScopeSeededProvider
from .inspect import build_kwargs_target
from .providers import AsyncCallableProviderImpl
from .providers import CallableProviderImpl
from .providers import ConstProviderImpl
from .providers import InternalProvider
from .providers import LinkProviderImpl
from .providers import ProviderImpl
from .scopes import ScopeSeededProviderImpl


##


PROVIDER_IMPLS_BY_PROVIDER: dict[type[Provider], ta.Callable[..., ProviderImpl]] = {
    AsyncFnProvider: lambda p: AsyncCallableProviderImpl(p, build_kwargs_target(p.fn)),
    FnProvider: lambda p: CallableProviderImpl(p, build_kwargs_target(p.fn)),
    CtorProvider: lambda p: CallableProviderImpl(p, build_kwargs_target(p.ty)),
    ConstProvider: ConstProviderImpl,
    LinkProvider: LinkProviderImpl,
    InternalProvider: lambda p: p.impl,

    ScopeSeededProvider: ScopeSeededProviderImpl,
}


def make_provider_impl(p: Provider) -> ProviderImpl:
    try:
        fac = PROVIDER_IMPLS_BY_PROVIDER[type(p)]
    except KeyError:
        pass
    else:
        return fac(p)

    raise TypeError(p)



"""
TODO:
 - check validity
 - signature vs getfullargspec - diff unwrapping + 'self' handling
"""
import dataclasses as dc
import enum
import inspect
import typing as ta

from ..lite.maybes import Maybe
from .classes.abstract import Abstract
from .classes.restrict import Final
from .classes.restrict import Sealed


T = ta.TypeVar('T')


CanParamSpec: ta.TypeAlias = ta.Union[
    'ParamSpec',
    inspect.Signature,
    ta.Callable,
]


##


@dc.dataclass(frozen=True, unsafe_hash=True)
class Param(Sealed, Abstract):
    name: str

    annotation: Maybe = Maybe.empty()

    prefix: ta.ClassVar[str] = ''

    @property
    def name_with_prefix(self) -> str:
        return f'{self.prefix}{self.name}'


#


@dc.dataclass(frozen=True, unsafe_hash=True)
class VarParam(Param, Abstract):
    pass


@dc.dataclass(frozen=True, unsafe_hash=True)
class ArgsParam(VarParam, Final):
    prefix: ta.ClassVar[str] = '*'


@dc.dataclass(frozen=True, unsafe_hash=True)
class KwargsParam(VarParam, Final):
    prefix: ta.ClassVar[str] = '**'


#


@dc.dataclass(frozen=True, unsafe_hash=True)
class ValParam(Param):
    default: Maybe = Maybe.empty()


@dc.dataclass(frozen=True, unsafe_hash=True)
class PosOnlyParam(ValParam, Final):
    pass


@dc.dataclass(frozen=True, unsafe_hash=True)
class KwOnlyParam(ValParam, Final):
    pass


#


class ParamSeparator(enum.Enum):
    POS_ONLY = '/'
    KW_ONLY = '*'


##


def _inspect_empty_to_maybe(o: T) -> Maybe[T]:
    if o is inspect.Parameter.empty:
        return Maybe.empty()
    else:
        return Maybe.just(o)


class ParamSpec(ta.Sequence[Param], Final):
    def __init__(self, *ps: Param) -> None:
        super().__init__()

        self._ps = ps

        self._hash: int | None = None

        self._has_annotations: bool | None = None
        self._has_defaults: bool | None = None

        self._with_seps: tuple[Param | ParamSeparator, ...] | None = None

    #

    @classmethod
    def of(cls, obj: CanParamSpec) -> 'ParamSpec':
        if isinstance(obj, ParamSpec):
            return obj
        elif isinstance(obj, inspect.Signature):
            return cls.of_signature(obj)
        else:
            return cls.inspect(obj)

    @classmethod
    def of_signature(
            cls,
            sig: inspect.Signature,
            *,
            offset: int = 0,
            strip_annotations: bool = False,
            strip_defaults: bool = False,
    ) -> 'ParamSpec':
        ps: list[Param] = []

        ip: inspect.Parameter
        for i, ip in enumerate(sig.parameters.values()):
            if i < offset:
                continue

            ann = _inspect_empty_to_maybe(ip.annotation) if not strip_annotations else Maybe.empty()
            dfl = _inspect_empty_to_maybe(ip.default) if not strip_defaults else Maybe.empty()

            if ip.kind == inspect.Parameter.POSITIONAL_ONLY:
                ps.append(PosOnlyParam(ip.name, ann, dfl))

            elif ip.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                ps.append(ValParam(ip.name, ann, dfl))

            elif ip.kind == inspect.Parameter.VAR_POSITIONAL:
                ps.append(ArgsParam(ip.name, ann))

            elif ip.kind == inspect.Parameter.KEYWORD_ONLY:
                ps.append(KwOnlyParam(ip.name, ann, dfl))

            elif ip.kind == inspect.Parameter.VAR_KEYWORD:
                ps.append(KwargsParam(ip.name, ann))

            else:
                raise ValueError(ip.kind)

        return cls(*ps)

    @classmethod
    def inspect(
            cls,
            obj: ta.Any,
            **kwargs: ta.Any,
    ) -> 'ParamSpec':
        return cls.of_signature(
            inspect.signature(obj),
            **kwargs,
        )

    #

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(map(repr, self._ps))})'

    def __hash__(self) -> int:
        if (h := self._hash) is not None:
            return h
        self._hash = h = hash(self._ps)
        return h

    def __eq__(self, other: object) -> bool:
        if type(other) is not ParamSpec:
            raise TypeError(other)
        return self._ps == other._ps  # noqa

    #

    @property
    def has_annotations(self) -> bool:
        if (ha := self._has_annotations) is not None:
            return ha
        self._has_annotations = ha = any(
            isinstance(p, (VarParam, ValParam)) and p.annotation.present
            for p in self._ps
        )
        return ha

    @property
    def has_defaults(self) -> bool:
        if (hd := self._has_defaults) is not None:
            return hd
        self._has_defaults = hd = any(
            isinstance(p, ValParam) and p.default.present
            for p in self._ps
        )
        return hd

    #

    @property
    def with_seps(self) -> ta.Sequence[Param | ParamSeparator]:
        if (ws := self._with_seps) is not None:
            return ws

        l: list[Param | ParamSeparator] = []
        needs_pos_only = False
        seen_kw_only = False
        for p in self._ps:
            if isinstance(p, PosOnlyParam):
                needs_pos_only = True
            elif needs_pos_only:
                l.append(ParamSeparator.POS_ONLY)
                needs_pos_only = False

            if isinstance(p, KwOnlyParam) and not seen_kw_only:
                l.append(ParamSeparator.KW_ONLY)
                seen_kw_only = True
            elif isinstance(p, KwargsParam):
                seen_kw_only = True

            l.append(p)

        self._with_seps = ws = tuple(l)
        return ws

    #

    @ta.overload
    def __getitem__(self, index: int) -> Param: ...

    @ta.overload
    def __getitem__(self, index: slice) -> ta.Sequence[Param]: ...

    def __getitem__(self, index):
        return self._ps[index]

    def __len__(self) -> int:
        return len(self._ps)


##


def param_render(
        p: Param | ParamSeparator,
        *,
        render_annotation: ta.Callable[[ta.Any], str] | None = None,
        render_default: ta.Callable[[ta.Any], str] | None = None,
) -> str:
    if isinstance(p, Param):
        ann_s: str | None = None
        if p.annotation.present:
            if render_annotation is None:
                raise ValueError(f'Param {p.name} has an annotation but no annotation renderer provided')
            ann_s = render_annotation(p.annotation.must())

        dfl_s: str | None = None
        if isinstance(p, ValParam) and p.default.present:
            if render_default is None:
                raise ValueError(f'Param {p.name} has a default but no default renderer provided')
            dfl_s = render_default(p.default.must())

        if ann_s is not None:
            if dfl_s is not None:
                return f'{p.name_with_prefix}: {ann_s} = {dfl_s}'
            else:
                return f'{p.name_with_prefix}: {ann_s}'
        elif dfl_s is not None:
            return f'{p.name_with_prefix}={dfl_s}'
        else:
            return p.name_with_prefix

    elif isinstance(p, ParamSeparator):
        return p.value

    else:
        raise TypeError(p)



# ruff: noqa: I001
from ... import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .flattening import (  # noqa
        ConfigFlattening as Flattening,
    )

    from .inheritance import (  # noqa
        build_config_inherited_values as build_inherited_values,
    )

    from .matching import (  # noqa
        MatchingConfigRewriter as MatchingRewriter,

        matched_config_rewrite as matched_rewrite,
    )

    from .merging import (  # noqa
        merge_configs as merge,
    )

    from .names import (  # noqa
        build_config_named_children as build_named_children,
    )

    from .rewriting import (  # noqa
        ConfigRewriterItem as RewriterItem,
        ConfigRewriterPath as RewriterPath,

        RawConfigMetadata as RawMetadata,

        ConfigRewriter as Rewriter,
    )

    from .strings import (  # noqa
        StringConfigRewriter as StringRewriter,

        format_config_strings as format_strings,
    )



# ruff: noqa: I001
import sys as _sys

from .. import lang as _lang


with _lang.auto_proxy_init(globals()) as _api_cap:
    ##

    ##
    # stdlib interface

    from dataclasses import (  # noqa
        FrozenInstanceError,

        MISSING,
        KW_ONLY,

        InitVar,
        Field,

        field,

        dataclass,
        make_dataclass,

        fields,

        is_dataclass,

        replace,
    )

    from .tools.as_ import (  # noqa
        asdict,
        astuple,
    )

    ##
    # additional interface

    from .impl.api.classes.decorator import (  # noqa
        dataclass as xdataclass,
    )

    from .impl.api.classes.make import (  # noqa
        make_dataclass as xmake_dataclass,
    )

    from .impl.api.classes.metadata import (  # noqa
        append_class_metadata,
        extra_class_params,
        init,
        metadata,
        validate,
    )

    from .impl.api.fields.metadata import (  # noqa
        extra_field_params,
        set_field_metadata,
        update_extra_field_params,
        with_extra_field_params,
    )

    from .impl.api.fields.constructor import (  # noqa
        field as xfield,
    )

    from .impl.concerns.replace import (  # noqa
        replace as xreplace,
    )

    from .impl.configs import (  # noqa
        init_package,
    )

    from .errors import (  # noqa
        FieldFnValidationError,
        FieldTypeValidationError,
        FieldValidationError,
        FnValidationError,
        TypeValidationError,
        ValidationError,
    )

    from .metaclass.bases import (  # noqa
        Box,
        Case,
        Data,
        Frozen,
    )

    from .metaclass.meta import (  # noqa
        DataMeta,
    )

    from .metaclass.specs import (  # noqa
        get_metaclass_spec,
    )

    from .reflection import (  # noqa
        reflect,
    )

    from .specs import (  # noqa
        CoerceFn,
        ValidateFn,
        ReprFn,

        InitFn,
        ClassValidateFn,

        DefaultFactory,

        FieldType,

        FieldSpec,

        ClassSpec,
    )

    from .tools.iter import (  # noqa
        fields_dict,

        iter_items,
        iter_keys,
        iter_values,
    )

    from .tools.modifiers import (  # noqa
        field_modifier,
        update_fields,
    )

    from .tools.only_ import (  # noqa
        only,
    )

    from .tools.replace import (  # noqa
        deep_replace,

        replace_if,
        replace_ne,
        replace_is_not,
    )

    from .tools.repr import (  # noqa
        opt_repr,
        truthy_repr,
    )

    from .tools.static import (  # noqa
        Static,
    )

    ##
    # lite imports

    from ..lite.dataclasses import (  # noqa
        dataclass_shallow_asdict as shallow_asdict,
        dataclass_shallow_astuple as shallow_astuple,

        is_immediate_dataclass,

        dataclass_maybe_post_init as maybe_post_init,

        dataclass_descriptor_method as descriptor_method,
    )


##
# globals hack

def _self_patching_global_proxy(l, r, ak_fn=None):
    def inner(*args, **kwargs):
        fn = getattr(_sys.modules[__name__], r)
        globals()[l] = fn
        if ak_fn:
            args, kwargs = ak_fn(*args, **kwargs)
        return fn(*args, **kwargs)
    return inner


globals()['field'] = _self_patching_global_proxy('field', 'xfield')

globals()['dataclass'] = _self_patching_global_proxy('dataclass', 'xdataclass')

globals()['make_dataclass'] = _self_patching_global_proxy(
    'make_dataclass',
    'xmake_dataclass',
    lambda *args, _frame_offset=1, **kwargs: (args, dict(_frame_offset=_frame_offset + 1, **kwargs)),
)

globals()['replace'] = _self_patching_global_proxy('replace', 'xreplace')


_api_cap.update_exports()
del _api_cap



r"""
https://github.com/edn-format/edn
https://github.com/antlr/grammars-v4/blob/master/edn/edn.g4
https://github.com/jorinvo/edn-data/blob/1e5824f63803eb58f35e98839352000053d47115/src/parse.ts
https://clojure.org/reference/reader#_extensible_data_notation_edn
"""
import dataclasses as dc
import io
import typing as ta

from ... import check
from ...funcs.genmachine import GenMachine


##


TokenKind: ta.TypeAlias = ta.Literal[
    'STRING',
    'CHAR',
    'WORD',
    'COMMENT',

    'LPAREN',
    'RPAREN',
    'LBRACKET',
    'RBRACKET',
    'HASH_LBRACE',
    'LBRACE',
    'RBRACE',

    'HASH_UNDERSCORE',
    'META',
    'QUOTE',

    'SPACE',
]


class Position(ta.NamedTuple):
    ofs: int
    line: int
    col: int


class Token(ta.NamedTuple):
    kind: TokenKind
    src: str

    pos: Position

    def __iter__(self):
        raise TypeError


##


SINGLE_TOKENS: ta.Mapping[str, TokenKind] = {
    '(': 'LPAREN',
    ')': 'RPAREN',
    '[': 'LBRACKET',
    ']': 'RBRACKET',
    '{': 'LBRACE',
    '}': 'RBRACE',

    '^': 'META',
    "'": 'QUOTE',
}


HASH_TOKENS: ta.Mapping[str, TokenKind] = {
    '{': 'HASH_LBRACE',
    '_': 'HASH_UNDERSCORE',
}


WORD_FIRST_SPECIAL_CHARS = ':.*+!-_?$%&=<>.'
WORD_BODY_SPECIAL_CHARS = '/'


##


@dc.dataclass()
class StreamLexError(Exception):
    message: str

    pos: Position


class StreamLexer(GenMachine[str, Token]):
    def __init__(
            self,
            *,
            include_space: bool = False,
    ) -> None:
        self._include_space = include_space

        self._ofs = 0
        self._line = 1
        self._col = 0

        self._buf = io.StringIO()

        super().__init__(self._do_main())

    @property
    def pos(self) -> Position:
        return Position(
            self._ofs,
            self._line,
            self._col,
        )

    def _char_in(self, c: str) -> str:
        if not isinstance(c, str):
            raise TypeError(c)
        if c and len(c) != 1:
            raise ValueError(c)

        self._ofs += 1

        if c == '\n':
            self._line += 1
            self._col = 0
        else:
            self._col += 1

        return c

    def _make_tok(
            self,
            kind: TokenKind,
            src: str,
            pos: Position,
    ) -> ta.Sequence[Token]:
        tok = Token(
            kind,
            src,
            pos,
        )
        return (tok,)

    def _flip_buf(self) -> str:
        src = self._buf.getvalue()
        self._buf.seek(0)
        self._buf.truncate()
        return src

    def _raise(self, msg: str, src: Exception | None = None) -> ta.NoReturn:
        raise StreamLexError(msg, self.pos) from src

    def _do_main(self, p: str | None = None):
        while True:
            if p is not None:
                c = p
                p = None
            else:
                c = self._char_in((yield None))  # noqa

            if not c:
                return None

            if c.isspace() or c == ',':
                if self._include_space:
                    yield self._make_tok('SPACE', c, self.pos)
                continue

            if c in SINGLE_TOKENS:
                yield self._make_tok(SINGLE_TOKENS[c], c, self.pos)
                continue

            if c == ';':
                return self._do_comment()

            if c == '"':
                return self._do_string()

            if c == '\\':
                return self._do_char()

            if c == '#':
                return self._do_hash()

            if (
                    c.isalnum() or
                    c in WORD_FIRST_SPECIAL_CHARS
            ):
                return self._do_word(c)

            self._raise(f'Unexpected input: {c}')

    def _do_comment(self):
        check.state(self._buf.tell() == 0)
        self._buf.write(';')

        pos = self.pos

        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c or c == '\n':
                break

            self._buf.write(c)

        src = self._flip_buf()
        yield self._make_tok('COMMENT', src, pos)
        return self._do_main()

    def _do_string(self):
        check.state(self._buf.tell() == 0)
        self._buf.write('"')

        pos = self.pos

        esc = False
        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c:
                self._raise(f'Unterminated string literal: {self._buf.getvalue()}')

            self._buf.write(c)
            if esc:
                esc = False
            elif c == '\\':
                esc = True
            elif c == '"':
                break

        src = self._flip_buf()
        yield self._make_tok('STRING', src, pos)
        return self._do_main()

    def _do_char(self):
        check.state(self._buf.tell() == 0)
        self._buf.write('\\')

        pos = self.pos

        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c or not (
                    c.isalnum() or
                    c == '\\'
            ):
                break

            self._buf.write(c)

        src = self._flip_buf()
        yield self._make_tok('CHAR', src, pos)
        return self._do_main(c)

    def _do_hash(self):
        check.state(self._buf.tell() == 0)

        pos = self.pos

        try:
            c = self._char_in((yield None))  # noqa
        except GeneratorExit:
            self._raise('Unexpected end of input')

        if (ht := HASH_TOKENS.get(c)) is not None:
            yield self._make_tok(ht, '#' + c, pos)
            return self._do_main()

        elif (
                c.isalnum() or
                c == '#' or
                c in WORD_FIRST_SPECIAL_CHARS
        ):
            return self._do_word('#' + c, pos=pos)

        else:
            self._raise(f'Unexpected input: {c}')

    def _do_word(self, pfx: str, *, pos: Position | None = None):
        check.state(self._buf.tell() == 0)
        self._buf.write(pfx)

        if pos is None:
            pos = self.pos

        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c or not (
                    c.isalnum() or
                    c in WORD_FIRST_SPECIAL_CHARS or
                    c in WORD_BODY_SPECIAL_CHARS
            ):
                break

            self._buf.write(c)

        src = self._flip_buf()
        yield self._make_tok('WORD', src, pos)
        return self._do_main(c)



"""
TODO:
 - jsonl pidfile
  - time started
  - code version / revision
  - venv?
"""
import abc
import threading
import typing as ta

from .. import cached
from .. import check
from .. import dataclasses as dc
from .. import lang
from ..configs.classes import Configurable
from .daemon import Daemon
from .targets import Target
from .targets import TargetRunner
from .targets import target_runner_for


ServiceT = ta.TypeVar('ServiceT', bound='Service')
ServiceConfigT = ta.TypeVar('ServiceConfigT', bound='Service.Config')


##


class Service(Configurable[ServiceConfigT], lang.Abstract):
    class Config(Configurable.Config, lang.Abstract):
        pass

    @classmethod
    def from_config(cls, config: Config) -> 'Service':
        return check.isinstance(config.configurable_cls(config), cls)

    #

    @abc.abstractmethod
    def _run(self) -> None:
        raise NotImplementedError

    def run(self) -> None:
        self._run()

    @classmethod
    def run_config(cls, config: Config) -> None:
        return cls.from_config(config).run()


##


class ServiceTarget(Target):
    svc: Service


class ServiceTargetRunner(TargetRunner, dc.Frozen):
    target: ServiceTarget

    def run(self) -> None:
        self.target.svc.run()


@target_runner_for.register
def _(target: ServiceTarget) -> ServiceTargetRunner:
    return ServiceTargetRunner(target)


#


class ServiceConfigTarget(Target):
    cfg: Service.Config


class ServiceConfigTargetRunner(TargetRunner, dc.Frozen):
    target: ServiceConfigTarget

    def run(self) -> None:
        Service.run_config(self.target.cfg)


@target_runner_for.register
def _(target: ServiceConfigTarget) -> ServiceConfigTargetRunner:
    return ServiceConfigTargetRunner(target)


##


@dc.dataclass(frozen=True)
class ServiceDaemon(lang.Final, ta.Generic[ServiceT, ServiceConfigT]):
    service: ServiceT | ServiceConfigT

    @cached.function
    def service_config(self) -> ServiceConfigT:
        with self._lock:
            if isinstance(self.service, Service):
                return self.service.config
            elif isinstance(self.service, Service.Config):
                return self.service
            else:
                raise TypeError(self.service)

    @cached.function
    def service_(self) -> ServiceT:
        with self._lock:
            if isinstance(self.service, Service):
                return self.service
            elif isinstance(self.service, Service.Config):
                return Service.from_config(self.service)  # type: ignore[return-value]
            else:
                raise TypeError(self.service)

    #

    daemon: Daemon | Daemon.Config = Daemon.Config()

    @cached.function
    def daemon_config(self) -> Daemon.Config:
        with self._lock:
            if isinstance(self.daemon, Daemon):
                return self.daemon.config
            elif isinstance(self.daemon, Daemon.Config):
                return self.daemon
            else:
                raise TypeError(self.daemon)

    @cached.function
    def daemon_(self) -> Daemon:
        with self._lock:
            if isinstance(self.daemon, Daemon):
                return self.daemon
            elif isinstance(self.daemon, Daemon.Config):
                return Daemon(Target.of(self.service_()), self.daemon)
            else:
                raise TypeError(self.daemon)

    #

    _lock: threading.RLock = dc.field(default_factory=lambda: threading.RLock(), init=False)



"""
While the core functionality lives in `lite/check.py` and is available in lite code, this module re-exports the contents
of its `check` API object as module-globals for convenience. It also sets up some non-lite integration, most notably
`ArgsRenderer` for rendering the AST's and values of failed checks (if the necessary optional dependencies are present).
"""
import typing as ta

from .lite.check import CheckArgsRenderer as ArgsRenderer  # noqa
from .lite.check import CheckExceptionFactory as ExceptionFactory  # noqa
from .lite.check import CheckLateConfigureFn as LateConfigureFn  # noqa
from .lite.check import CheckMessage as Message  # noqa
from .lite.check import CheckOnRaiseFn as OnRaiseFn  # noqa
from .lite.check import Checks
from .lite.check import check


T = ta.TypeVar('T')
SizedT = ta.TypeVar('SizedT', bound=ta.Sized)


_isinstance = isinstance
_issubclass = issubclass
_callable = callable


##


try:
    from . import _check  # type: ignore
except ImportError:
    pass
else:
    setattr(Checks, '_unpack_isinstance_spec', _check.unpack_isinstance_spec)


##


def register_on_raise(fn: OnRaiseFn) -> None:
    check.register_on_raise(fn)


def unregister_on_raise(fn: OnRaiseFn) -> None:
    check.unregister_on_raise(fn)


##


def _try_get_args_rendering() -> ArgsRenderer | None:
    try:
        from .diag.asts import ArgsRenderer

        ArgsRenderer.smoketest()

    except Exception:  # noqa
        return None

    def _real_render_args(fmt: str, *args: ta.Any) -> str | None:
        ra = ArgsRenderer(back=3).render_args(*args)
        if ra is None:
            return None

        return fmt % tuple(str(a) for a in ra)

    return _real_render_args


def _try_enable_args_rendering(c: Checks) -> None:
    if (rf := _try_get_args_rendering()) is not None:
        c.set_args_renderer(rf)


check.register_late_configure(_try_enable_args_rendering)


##
# The following code manually proxies to the lite code, as opposed to simply assigning global names to instance methods
# of `check`, to assist type analysis in tools - pycharm for example has a hard time deducing the return types in such
# cases. The functions are however dynamically overwritten below with direct references to the method to remove one
# layer of call indirection at runtime.


_CHECK_PROXY_FUNCTIONS: dict[str, ta.Any] = {}


def _check_proxy_function(fn: T) -> T:
    name = fn.__name__  # type: ignore[attr-defined]
    if name in _CHECK_PROXY_FUNCTIONS:
        raise NameError(fn)
    _CHECK_PROXY_FUNCTIONS[name] = fn
    return fn


##


@ta.overload
def isinstance(v: ta.Any, spec: type[T], msg: Message = None) -> T:  # noqa
    ...


@ta.overload
def isinstance(v: ta.Any, spec: ta.Any, msg: Message = None) -> ta.Any:  # noqa
    ...


@_check_proxy_function
def isinstance(v, spec, msg=None):  # noqa
    return check.isinstance(v, spec, msg=msg)


@ta.overload
def of_isinstance(spec: type[T], msg: Message = None) -> ta.Callable[[ta.Any], T]:
    ...


@ta.overload
def of_isinstance(spec: ta.Any, msg: Message = None) -> ta.Callable[[ta.Any], ta.Any]:
    ...


@_check_proxy_function
def of_isinstance(spec, msg=None):
    return check.of_isinstance(spec, msg=msg)


@_check_proxy_function
def cast(v: ta.Any, cls: type[T], msg: Message = None) -> T:
    return check.cast(v, cls, msg=msg)


@_check_proxy_function
def of_cast(cls: type[T], msg: Message = None) -> ta.Callable[[T], T]:
    return check.of_cast(cls, msg=msg)


@_check_proxy_function
def not_isinstance(v: T, spec: ta.Any, msg: Message = None) -> T:
    return check.not_isinstance(v, spec, msg=msg)


@_check_proxy_function
def of_not_isinstance(spec: ta.Any, msg: Message = None) -> ta.Callable[[T], T]:
    return check.of_not_isinstance(spec, msg=msg)


#


@_check_proxy_function
def issubclass(v: type[T], spec: ta.Any, msg: Message = None) -> type[T]:  # noqa
    return check.issubclass(v, spec, msg=msg)


@_check_proxy_function
def not_issubclass(v: type[T], spec: ta.Any, msg: Message = None) -> type[T]:
    return check.not_issubclass(v, spec, msg=msg)


#


@_check_proxy_function
def in_(v: T, c: ta.Container[T], msg: Message = None) -> T:
    return check.in_(v, c, msg=msg)


@_check_proxy_function
def not_in(v: T, c: ta.Container[T], msg: Message = None) -> T:
    return check.not_in(v, c, msg=msg)


@_check_proxy_function
def empty(v: SizedT, msg: Message = None) -> SizedT:
    return check.empty(v, msg=msg)


@_check_proxy_function
def iterempty(v: ta.Iterable[T], msg: Message = None) -> ta.Iterable[T]:
    return check.iterempty(v, msg=msg)


@_check_proxy_function
def not_empty(v: SizedT, msg: Message = None) -> SizedT:
    return check.not_empty(v, msg=msg)


@_check_proxy_function
def unique(it: ta.Iterable[T], msg: Message = None) -> ta.Iterable[T]:
    return check.unique(it, msg=msg)


@_check_proxy_function
def single(obj: ta.Iterable[T], msg: Message = None) -> T:
    return check.single(obj, msg=msg)


@_check_proxy_function
def opt_single(obj: ta.Iterable[T], msg: Message = None) -> T | None:
    return check.opt_single(obj, msg=msg)


#


@_check_proxy_function
def none(v: ta.Any, msg: Message = None) -> None:
    return check.none(v, msg=msg)


@_check_proxy_function
def not_none(v: T | None, msg: Message = None) -> T:
    return check.not_none(v, msg=msg)


#


@_check_proxy_function
def equal(v: T, o: ta.Any, msg: Message = None) -> T:
    return check.equal(v, o, msg=msg)


@_check_proxy_function
def not_equal(v: T, o: ta.Any, msg: Message = None) -> T:
    return check.not_equal(v, o, msg=msg)


@_check_proxy_function
def is_(v: T, o: ta.Any, msg: Message = None) -> T:
    return check.is_(v, o, msg=msg)


@_check_proxy_function
def is_not(v: T, o: ta.Any, msg: Message = None) -> T:
    return check.is_not(v, o, msg=msg)


@_check_proxy_function
def callable(v: T, msg: Message = None) -> T:  # noqa
    return check.callable(v, msg=msg)


@_check_proxy_function
def non_empty_str(v: str | None, msg: Message = None) -> str:
    return check.non_empty_str(v, msg=msg)


@_check_proxy_function
def replacing(expected: ta.Any, old: ta.Any, new: T, msg: Message = None) -> T:
    return check.replacing(expected, old, new, msg=msg)


@_check_proxy_function
def replacing_none(old: ta.Any, new: T, msg: Message = None) -> T:
    return check.replacing_none(old, new, msg=msg)


#


@_check_proxy_function
def arg(v: bool, msg: Message = None) -> None:
    return check.arg(v, msg=msg)


@_check_proxy_function
def state(v: bool, msg: Message = None) -> None:
    return check.state(v, msg=msg)


##


def _install_direct_check_proxy_functions() -> None:
    for n in _CHECK_PROXY_FUNCTIONS:
        globals()[n] = getattr(check, n)


_install_direct_check_proxy_functions()



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
Should be kept somewhat lightweight - used in cli entrypoints.

TODO:
 - real relative cls names - shouldn't need parent package names
 - TypeMap style weak cache of issubclass queries
  - wait.. lazily load the class for virtual subclass queries? xor support virtual bases?
 - weakref class dict keys?
 - cheap_discover_package_root_dirs: Seq[str] | None = None or smth
 - maaaybe.. add an EnvVar? OMLISH_MANIFEST_ROOT_DIRS? if set to : delimited, turns off package disco and overrides
   scan_root_dirs
  - currently the cli cant subprocess itself and keep manifests working
  - EnvVar cls is already lite
 - can discover_packages deadlock with concurrent / multithreaded imports?
"""
import dataclasses as dc
import importlib.machinery
import importlib.util
import json
import os.path
import threading
import typing as ta

from .types import Manifest


T = ta.TypeVar('T')


##


class ManifestLoader:
    class LoadedManifest:
        def __init__(
                self,
                package: 'ManifestLoader.LoadedPackage',
                resolved: 'ManifestLoader._ResolvedManifest',
        ) -> None:
            super().__init__()

            self._package = package
            self._resolved = resolved

        def __repr__(self) -> str:
            return (
                f'{self.__class__.__name__}@{id(self):x}('
                f'package={self._package.name!r}, '
                f'module={self._resolved.module!r}, '
                f'class_key={self._resolved.class_key!r}'
                f')'
            )

        @property
        def package(self) -> 'ManifestLoader.LoadedPackage':
            return self._package

        @property
        def module(self) -> str:
            return self._resolved.module

        @property
        def class_key(self) -> str:
            return self._resolved.class_key

        @property
        def manifest(self) -> Manifest:
            return self._resolved.manifest

        @property
        def loader(self) -> 'ManifestLoader':
            return self._package.loader

        _value: ta.Any

        def value(self) -> ta.Any:
            try:
                return self._value
            except AttributeError:
                pass

            value = self.loader._instantiate_resolved_manifest(self._resolved)  # noqa
            self._value = value
            return value

    class LoadedPackage:
        def __init__(
                self,
                loader: 'ManifestLoader',
                name: str,
        ) -> None:
            super().__init__()

            self._loader = loader
            self._name = name

        _manifests: ta.Sequence['ManifestLoader.LoadedManifest']

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}@{id(self):x}(name={self._name!r})'

        @property
        def loader(self) -> 'ManifestLoader':
            return self._loader

        @property
        def name(self) -> str:
            return self._name

        @property
        def manifests(self) -> ta.Sequence['ManifestLoader.LoadedManifest']:
            return self._manifests

        _manifests_by_class_key: ta.Mapping[str, ta.Sequence['ManifestLoader.LoadedManifest']]

        @property
        def manifests_by_class_key(self) -> ta.Mapping[str, ta.Sequence['ManifestLoader.LoadedManifest']]:
            try:
                return self._manifests_by_class_key
            except AttributeError:
                pass

            dct: dict = {}
            for m in self._manifests:
                try:
                    lst = dct[m.class_key]
                except KeyError:
                    lst = dct[m.class_key] = []
                lst.append(m)
            self._manifests_by_class_key = dct
            return dct

    ##

    @dc.dataclass(frozen=True)
    class Config:
        package_scan_root_dirs: ta.Optional[ta.Collection[str]] = None

        discover_packages: ta.Optional[bool] = None
        discover_packages_fallback_scan_root_dirs: ta.Optional[ta.Collection[str]] = None

        module_remap: ta.Optional[ta.Mapping[str, str]] = None

        value_instantiator: ta.Optional[ta.Callable[..., ta.Any]] = None

        def __post_init__(self) -> None:
            if isinstance(self.package_scan_root_dirs, str):
                raise TypeError(self.package_scan_root_dirs)
            if isinstance(self.discover_packages_fallback_scan_root_dirs, bool):
                raise TypeError(self.discover_packages_fallback_scan_root_dirs)

        @classmethod
        def merge(cls, *configs: 'ManifestLoader.Config') -> 'ManifestLoader.Config':
            kw: dict = {}
            for c in configs:
                for k, v in dc.asdict(c).items():
                    if v is None:
                        continue
                    elif k in ('package_scan_root_dirs', 'discover_packages_fallback_scan_root_dirs'):  # noqa
                        kw[k] = [*kw.get(k, []), *v]
                    elif k == 'module_remap':
                        kw[k] = {**kw.get(k, {}), **v}
                    else:
                        kw[k] = v
            return cls(**kw)

        def __or__(self, other: 'ManifestLoader.Config') -> 'ManifestLoader.Config':
            return ManifestLoader.Config.merge(self, other)

    def __init__(
            self,
            config: Config,
    ) -> None:
        super().__init__()

        self._config = config

        self._lock = threading.RLock()

        self._module_remap = config.module_remap or {}
        self._module_reverse_remap = {v: k for k, v in (self._module_remap or {}).items()}

        self._loaded_classes: ta.Dict[str, type] = {}
        self._loaded_packages: ta.Dict[str, ta.Optional[ManifestLoader.LoadedPackage]] = {}

        self._scanned_package_root_dirs: ta.Dict[str, ta.Sequence[str]] = {}

    @property
    def config(self) -> Config:
        return self._config

    @classmethod
    def config_from_entry_point(
            cls,
            globals: ta.Mapping[str, ta.Any],  # noqa
    ) -> Config:
        rm: ta.Dict[str, str] = {}

        if '__name__' in globals and '__spec__' in globals:
            name: str = globals['__name__']
            spec: importlib.machinery.ModuleSpec = globals['__spec__']
            if '__main__' not in rm and name == '__main__':
                rm[spec.name] = '__main__'

        return ManifestLoader.Config(module_remap=rm)

    ##

    ENTRY_POINT_GROUP: ta.ClassVar[str] = 'omlish.manifests'

    _discovered_packages: ta.ClassVar[ta.Optional[ta.Sequence[str]]] = None

    @classmethod
    def _discover_packages_uncached(cls) -> ta.Sequence[str]:
        from importlib import metadata as importlib_metadata  # noqa

        return [
            ep.value
            for ep in importlib_metadata.entry_points(group=cls.ENTRY_POINT_GROUP)
        ]

    @classmethod
    def discover_packages(cls) -> ta.Sequence[str]:
        if (x := cls._discovered_packages) is not None:
            return x

        x = cls._discover_packages_uncached()
        cls._discovered_packages = x
        return x

    ##

    def _scan_package_root_dir_uncached(
            self,
            root_dir: str,
    ) -> ta.Sequence[str]:
        pkgs: ta.List[str] = []

        for n in os.listdir(root_dir):
            if (
                    os.path.isdir(p := os.path.join(root_dir, n)) and
                    os.path.exists(os.path.join(p, '__init__.py'))
            ):
                pkgs.append(n)

        return pkgs

    def _scan_package_root_dir_locked(
            self,
            root_dir: str,
    ) -> ta.Sequence[str]:
        try:
            return self._scanned_package_root_dirs[root_dir]
        except KeyError:
            pass

        ret = self._scan_package_root_dir_uncached(root_dir)
        self._scanned_package_root_dirs[root_dir] = ret
        return ret

    def _scan_package_root_dir(
            self,
            root_dir: str,
    ) -> ta.Sequence[str]:
        with self._lock:
            return self._scan_package_root_dir_locked(root_dir)

    ##

    def _detect_packages_uncached(self) -> ta.AbstractSet[str]:
        ret: ta.Set[str] = set()

        for r in self._config.package_scan_root_dirs or []:
            ret.update(self._scan_package_root_dir_locked(r))

        if self._config.discover_packages:
            ret.update(dps := self.discover_packages())
            if not dps:
                for r in self._config.discover_packages_fallback_scan_root_dirs or []:
                    ret.update(self._scan_package_root_dir_locked(r))

        return ret

    _detected_packages: ta.Optional[ta.AbstractSet[str]] = None

    def _detect_packages_locked(self) -> ta.AbstractSet[str]:
        if self._detected_packages is None:
            self._detected_packages = self._detect_packages_uncached()

        return self._detected_packages

    def detect_packages(self) -> ta.AbstractSet[str]:
        with self._lock:
            return self._detect_packages_locked()

    ##

    @dc.dataclass()
    class ClassKeyError(Exception):
        key: str
        module: ta.Optional[str] = None

    def _load_class_uncached(self, key: str) -> type:
        if not key.startswith('!'):
            raise ManifestLoader.ClassKeyError(key)

        parts = key[1:].split('.')
        if '' in parts:
            raise ManifestLoader.ClassKeyError(key)

        pos = next(i for i, p in enumerate(parts) if p[0].isupper())

        mod_name = '.'.join(parts[:pos])
        mod_name = self._module_remap.get(mod_name, mod_name)
        mod = importlib.import_module(mod_name)

        obj: ta.Any = mod
        for ca in parts[pos:]:
            obj = getattr(obj, ca)

        cls = obj
        if not isinstance(cls, type):
            raise TypeError(cls)

        return cls

    def _load_class_locked(self, key: str) -> type:
        try:
            return self._loaded_classes[key]
        except KeyError:
            pass

        cls = self._load_class_uncached(key)
        self._loaded_classes[key] = cls
        return cls

    def _load_class(self, key: str) -> type:
        with self._lock:
            return self._load_class_locked(key)

    def get_class_key(self, cls: type) -> str:
        if not (isinstance(cls, type) and dc.is_dataclass(cls)):
            raise TypeError(cls)
        mod_name = cls.__module__
        mod_name = self._module_reverse_remap.get(mod_name, mod_name)
        return f'!{mod_name}.{cls.__qualname__}'

    ##

    class _ResolvedManifest(ta.NamedTuple):
        manifest: Manifest
        package: str

        module: str
        class_key: str
        value_dct: ta.Any

    @classmethod
    def _resolve_raw_manifest(
            cls,
            m: Manifest,
            *,
            package_name: str,
    ) -> _ResolvedManifest:
        # self._module = module
        # self._class_key = class_key
        if not m.module.startswith('.'):
            raise NameError(m.module)

        module = package_name + m.module

        [(class_key, value_dct)] = m.value.items()

        if not class_key.startswith('!'):
            raise ManifestLoader.ClassKeyError(class_key, module=module)

        if class_key.startswith('!.'):
            class_key = f'!{package_name}{class_key[1:]}'

        # FIXME: move to builder
        # elif key.startswith('$.'):
        #     if module.startswith('.'):
        #         raise NameError(module)
        #     kl = key[1:].split('.')
        #     ml = module.split('.')
        #     lvl = len(kl) - kl[::-1].index('')
        #     if lvl >= len(ml):
        #         raise ManifestLoader.ClassKeyError(key, module=module)
        #     rn = '.'.join([*ml[:-lvl], *kl[lvl:]])
        #     return f'${rn}'

        return ManifestLoader._ResolvedManifest(
            manifest=m,
            package=package_name,

            module=module,
            class_key=class_key,
            value_dct=value_dct,
        )

    ##

    @classmethod
    def _deserialize_raw_manifests(cls, obj: ta.Any) -> ta.Sequence[Manifest]:
        if not isinstance(obj, (list, tuple)):
            raise TypeError(obj)

        lst: ta.List[Manifest] = []
        for e in obj:
            lst.append(Manifest(**e))

        return lst

    @classmethod
    def _read_package_file_text(cls, package_name: str, file_name: str) -> ta.Optional[str]:
        # importlib.resources.files actually imports the package - to avoid this, if possible, the file is read straight
        # off the filesystem.
        # FIXME: find_spec *still* imports the parent package to get __path__ to feed to _find_spec...
        spec = importlib.util.find_spec(package_name)
        if (
                spec is not None and
                isinstance(spec.loader, importlib.machinery.SourceFileLoader) and
                spec.origin is not None and
                len(spec.submodule_search_locations or []) == 1 and
                os.path.basename(spec.origin) == '__init__.py' and
                os.path.isfile(spec.origin)
        ):
            file_path = os.path.join(os.path.dirname(spec.origin), file_name)
            if os.path.isfile(file_path):
                with open(file_path, encoding='utf-8') as f:
                    return f.read()

        from importlib import resources as importlib_resources  # noqa

        t = importlib_resources.files(package_name).joinpath(file_name)
        if not t.is_file():
            return None
        return t.read_text('utf-8')

    MANIFESTS_FILE_NAME: ta.ClassVar[str] = '.omlish-manifests.json'

    @classmethod
    def _read_package_raw_manifests(cls, package_name: str) -> ta.Optional[ta.Sequence[Manifest]]:
        src = cls._read_package_file_text(package_name, cls.MANIFESTS_FILE_NAME)
        if src is None:
            return None

        obj = json.loads(src)
        if not isinstance(obj, (list, tuple)):
            raise TypeError(obj)

        return cls._deserialize_raw_manifests(obj)

    ##

    def _load_package_uncached(self, package_name: str) -> ta.Optional[LoadedPackage]:
        if (raw_lst := self._read_package_raw_manifests(package_name)) is None:
            return None

        ld_pkg = ManifestLoader.LoadedPackage(self, package_name)

        ld_man_lst: ta.List[ManifestLoader.LoadedManifest] = []
        for raw in raw_lst:
            rsv = self._resolve_raw_manifest(raw, package_name=package_name)

            ld_man = ManifestLoader.LoadedManifest(ld_pkg, rsv)

            ld_man_lst.append(ld_man)

        ld_pkg._manifests = ld_man_lst  # noqa

        return ld_pkg

    def _load_package_locked(self, package_name: str) -> ta.Optional[LoadedPackage]:
        try:
            return self._loaded_packages[package_name]
        except KeyError:
            pass

        pkg = self._load_package_uncached(package_name)
        self._loaded_packages[package_name] = pkg
        return pkg

    def load_package(self, package_name: str) -> ta.Optional[LoadedPackage]:
        with self._lock:
            return self._load_package_locked(package_name)

    ##

    def _instantiate_value(self, cls: type, **kwargs: ta.Any) -> ta.Any:
        if self._config.value_instantiator is not None:
            return self._config.value_instantiator(cls, **kwargs)
        else:
            return cls(**kwargs)

    def _instantiate_resolved_manifest(self, resolved: _ResolvedManifest) -> ta.Any:
        cls = self._load_class(resolved.class_key)
        value = self._instantiate_value(cls, **resolved.value_dct)
        return value

    ##

    def _load_locked(
            self,
            *,
            packages: ta.Optional[ta.Collection[str]] = None,
            classes: ta.Optional[ta.Collection[type]] = None,
    ) -> ta.Sequence[LoadedManifest]:
        if isinstance(packages, str):
            raise TypeError(packages)

        class_keys: ta.Optional[ta.Set[str]] = None
        if classes is not None:
            class_keys = {self.get_class_key(cls) for cls in classes}

        if packages is None:
            packages = self._detect_packages_locked()

        lst: ta.List[ManifestLoader.LoadedManifest] = []
        for pn in packages:
            if (lp := self._load_package_locked(pn)) is None:
                continue

            if class_keys is not None:
                for ck in class_keys:
                    lst.extend(lp.manifests_by_class_key.get(ck, []))

            else:
                lst.extend(lp.manifests)

        return lst

    def load(
            self,
            *,
            packages: ta.Optional[ta.Collection[str]] = None,
            classes: ta.Optional[ta.Collection[type]] = None,
    ) -> ta.Sequence[LoadedManifest]:
        if isinstance(packages, str):
            raise TypeError(packages)

        with self._lock:
            return self._load_locked(
                packages=packages,
                classes=classes,
            )

    def load_values(
            self,
            *,
            packages: ta.Optional[ta.Collection[str]] = None,
            classes: ta.Optional[ta.Collection[type]] = None,
    ) -> ta.Sequence[ta.Any]:
        return [
            lm.value()
            for lm in self.load(
                packages=packages,
                classes=classes,
            )
        ]

    def load_values_of(
            self,
            cls: ta.Type[T],
            *,
            packages: ta.Optional[ta.Collection[str]] = None,
    ) -> ta.Sequence[T]:
        return [
            ta.cast(T, lm.value())
            for lm in self.load(
                packages=packages,
                classes=[cls],
            )
        ]



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from ... import reflect as rfl
from .types import NotSpecified


##


_NOT_SPECIFIED_RTY = rfl.type_(type[NotSpecified])


@dc.dataclass(frozen=True)
class NotSpecifiedUnionMarshaler(msh.Marshaler):
    m: msh.Marshaler

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        if o is NotSpecified:
            raise TypeError(o)
        return self.m.marshal(ctx, o)


class NotSpecifiedUnionMarshalerFactory(msh.MarshalerFactory):
    def make_marshaler(self, ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Marshaler] | None:
        if not (
            isinstance(rty, rfl.Union) and
            not rty.is_optional and
            _NOT_SPECIFIED_RTY in rty.args
        ):
            return None

        def inner() -> msh.Marshaler:
            args = set(check.isinstance(rty, rfl.Union).args) - {_NOT_SPECIFIED_RTY}
            nty = rfl.type_(ta.Union[*args])
            m = ctx.make_marshaler(nty)
            return NotSpecifiedUnionMarshaler(m)

        return inner


class NotSpecifiedUnionUnmarshalerFactory(msh.UnmarshalerFactory):
    def make_unmarshaler(self, ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Unmarshaler] | None:  # noqa
        if not (
            isinstance(rty, rfl.Union) and
            not rty.is_optional and
            _NOT_SPECIFIED_RTY in rty.args
        ):
            return None

        def inner() -> msh.Unmarshaler:
            args = set(check.isinstance(rty, rfl.Union).args) - {_NOT_SPECIFIED_RTY}
            nty = rfl.type_(ta.Union[*args])
            return ctx.make_unmarshaler(nty)

        return inner


@lang.static_init
def _install_standard_marshaling() -> None:
    msh.install_standard_factories(
        msh.ForbiddenTypeMarshalerFactory({_NOT_SPECIFIED_RTY}),
        msh.ForbiddenTypeUnmarshalerFactory({_NOT_SPECIFIED_RTY}),

        NotSpecifiedUnionMarshalerFactory(),
        NotSpecifiedUnionUnmarshalerFactory(),
    )



import io
import textwrap
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import dispatch
from .. import lang
from .content import Dom
from .content import iter_content


if ta.TYPE_CHECKING:
    import markupsafe as ms

    _HAS_MARKUPSAFE = True

else:
    ms = lang.proxy_import('markupsafe')

    _HAS_MARKUPSAFE = lang.can_import('markupsafe')


##


@dc.dataclass()
class InvalidTagError(Exception):
    pass


@dc.dataclass()
class StrForbiddenError(Exception):
    pass


class Renderer:
    def __init__(
            self,
            sb: io.StringIO | None,
            *,
            forbid_str: bool = False,
            escape: ta.Callable[[str], str] | None = None,
            indent: str | int | None = None,
            indent_string_content: bool = False,
    ) -> None:
        super().__init__()

        if sb is None:
            sb = io.StringIO()
        self._sb = sb

        self._forbid_str = forbid_str
        self._escape_fn = escape
        if isinstance(indent, int):
            indent = ' ' * indent
        self._indent_unit: str | None = indent
        self._indent_string_content = indent_string_content

        self._level = 0
        self._indent_cache: dict[int, str] = {}

    @classmethod
    def render_to_str(cls, o: ta.Any, **kwargs: ta.Any) -> str:
        sb = io.StringIO()
        cls(sb, **kwargs).render(o)
        return sb.getvalue()

    #

    def _escape(self, s: str) -> str:
        if _HAS_MARKUPSAFE and isinstance(s, ms.Markup):
            return s
        if type(s) is not str:
            raise TypeError(s)
        if (fn := self._escape_fn) is None:
            return s
        return fn(s)

    def _check_tag(self, s: str) -> str:
        if self._escape(s) != s:
            raise InvalidTagError(s)
        return s

    #

    def _indent_str(self) -> str | None:
        if not (s := self._indent_unit):
            return None
        try:
            return self._indent_cache[self._level]
        except KeyError:
            pass
        ls = self._indent_cache[self._level] = s * self._level
        return ls

    def _write_indent(self) -> None:
        if (s := self._indent_str()):
            self._sb.write(s)

    #

    @dispatch.method(instance_cache=True)
    def render(self, o: ta.Any) -> None:
        raise TypeError(o)

    #

    def _write_string_content(self, s: str) -> None:
        if self._indent_string_content and (ls := self._indent_str()):
            s = textwrap.indent(s, ls)
        self._sb.write(s)

    @render.register  # noqa
    def _render_str(self, s: str) -> None:
        if self._forbid_str:
            raise StrForbiddenError(s)
        self._write_string_content(s)

    @render.register  # noqa
    def _render_sequence(self, l: ta.Sequence) -> None:
        for e in l:
            self.render(e)

    #

    if _HAS_MARKUPSAFE:
        @render.register
        def _render_markup(self, m: ms.Markup) -> None:
            self._write_string_content(m)

    #

    def _render_tag(self, tag: str) -> None:
        self._sb.write(self._check_tag(tag))

    def _render_attr_key(self, k: str) -> None:
        self._sb.write(self._check_tag(k))

    def _render_attr_value(self, v: ta.Any) -> None:
        self._sb.write('"')
        self._sb.write(check.isinstance(v, str))  # FIXME
        self._sb.write('"')

    @render.register  # noqa
    def _render_dom(self, n: Dom) -> None:
        self._write_indent()

        self._sb.write('<')
        self._render_tag(n.tag)

        for k, v in (n.attrs or {}).items():
            self._sb.write(' ')
            self._render_attr_key(k)
            if v is not None:
                self._sb.write('=')
                self._render_attr_value(v)

        i = -1
        for i, c in enumerate(iter_content(n.body)):
            if not i:
                self._sb.write('>')

            if self._indent_unit:
                self._sb.write('\n')

            self._level += 1
            self.render(c)
            self._level -= 1

        if i >= 0:
            if self._indent_unit:
                self._sb.write('\n')
                self._write_indent()

            self._sb.write('</')
            self._render_tag(n.tag)
            self._sb.write('>')

        else:
            self._sb.write(' />')


render = Renderer.render_to_str



import typing as ta

from ... import dataclasses as dc
from ... import lang
from .base import Builder
from .selects import Select
from .stmts import Stmt


##


class Union(Stmt, lang.Final):
    selects: ta.Sequence[Select] = dc.xfield(coerce=tuple)

    _: dc.KW_ONLY

    all: bool = False


class UnionBuilder(Builder):
    def union(self, *selects: Select, all: bool = False) -> Union:  # noqa
        return Union(selects, all=all)



# ruff: noqa: UP006 UP007 UP045 UP046
# @omlish-lite
import abc
import typing as ta

from ...lite.abstract import Abstract
from ...lite.reflect import is_union_alias
from ...lite.strings import snake_case
from ...lite.wrappers import update_wrapper_no_annotations


if ta.TYPE_CHECKING:
    from .bindings import TypedLoggerBindings
    from .contexts import TypedLoggerContext


T = ta.TypeVar('T')
U = ta.TypeVar('U')

TypedLoggerValueT = ta.TypeVar('TypedLoggerValueT', bound='TypedLoggerValue')

TypedLoggerValueOrProvider = ta.Union['TypedLoggerValue', 'TypedLoggerValueProvider']  # ta.TypeAlias

AbsentTypedLoggerValue = ta.Type['ABSENT_TYPED_LOGGER_VALUE']  # ta.TypeAlias

TypedLoggerValueOrAbsent = ta.Union['TypedLoggerValue', AbsentTypedLoggerValue]
TypedLoggerValueOrProviderOrAbsent = ta.Union[TypedLoggerValueOrProvider, AbsentTypedLoggerValue]  # ta.TypeAlias

TypedLoggerFieldValue = ta.Union[TypedLoggerValueOrProviderOrAbsent, ta.Type['TypedLoggerValue'], 'TypedLoggerConstFieldValue']  # ta.TypeAlias  # noqa
ResolvedTypedLoggerFieldValue = ta.Union[TypedLoggerValueOrAbsent, 'TypedLoggerConstFieldValue']  # ta.TypeAlias
UnwrappedTypedLoggerFieldValue = ta.Union[TypedLoggerValueOrAbsent, ta.Any]  # ta.TypeAlias


##


class _TypedLoggerInternalMethods(ta.Protocol):  # noqa
    def _typed_logger_provide_value(self, ctx: 'TypedLoggerContext') -> TypedLoggerValueOrAbsent: ...  # noqa

    def _typed_logger_maybe_provide_default_value(self, ctx: 'TypedLoggerContext') -> ta.Tuple[TypedLoggerValueOrAbsent, ...]: ... # noqa

    def _typed_logger_resolve_field_value(self, ctx: 'TypedLoggerContext') -> ResolvedTypedLoggerFieldValue: ...  # noqa

    def _typed_logger_unwrap_field_value(self, ctx: 'TypedLoggerContext') -> UnwrappedTypedLoggerFieldValue: ...  # noqa

    def _typed_logger_visit_bindings(self, vst: 'TypedLoggerBindings._Visitor') -> None:  ...  # noqa


##


@ta.final
class ABSENT_TYPED_LOGGER_VALUE:  # noqa
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    #

    @classmethod
    def map_absent(cls, fn: ta.Callable[[TypedLoggerValueT], U]) -> AbsentTypedLoggerValue:  # noqa
        return cls

    #

    @classmethod
    def _typed_logger_provide_value(cls, ctx: 'TypedLoggerContext') -> TypedLoggerValueOrAbsent:  # noqa
        return cls

    @classmethod
    def _typed_logger_resolve_field_value(cls, ctx: 'TypedLoggerContext') -> ResolvedTypedLoggerFieldValue:  # noqa
        return cls

    @classmethod
    def _typed_logger_unwrap_field_value(cls, ctx: 'TypedLoggerContext') -> UnwrappedTypedLoggerFieldValue:  # noqa
        return cls


##


class TypedLoggerValue(Abstract, ta.Generic[T]):
    def __init__(self, v: T) -> None:
        self._v = v

    @classmethod
    def of(cls: ta.Type[TypedLoggerValueT], v: ta.Any) -> ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]:
        return cls(v) if v is not ABSENT_TYPED_LOGGER_VALUE else ABSENT_TYPED_LOGGER_VALUE

    @ta.final
    @property
    def v(self) -> T:
        return self._v

    def __repr__(self) -> str:
        return f'{self.__class__.__module__}.{self.__class__.__qualname__}({self._v!r})'

    #

    def map_absent(  # noqa
            self: TypedLoggerValueT,
            fn: ta.Callable[[TypedLoggerValueT], U],
    ) -> ta.Union[U, AbsentTypedLoggerValue]:
        return fn(self)

    #

    _hash: int

    def __hash__(self) -> int:
        try:
            return self._hash
        except AttributeError:
            pass
        h = self._hash = hash((self.__class__, self.v))
        return h

    def __eq__(self, o):
        if not isinstance(o, TypedLoggerValue):
            return NotImplemented
        return self.__class__ is o.__class__ and self.v == o.v

    def __ne__(self, o):
        return not (self == o)

    #

    _default_key: ta.ClassVar[ta.Union[str, bool]] = False

    @ta.final
    @classmethod
    def default_key(cls) -> ta.Optional[str]:
        return cls.__default_key  # type: ignore[attr-defined]

    #

    class ContextLambda(ta.NamedTuple):
        fn: ta.Callable[['TypedLoggerContext'], ta.Any]

    #

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if '__call__' in cls.__dict__:
            raise TypeError(f'{cls} should not be callable')

        try:
            dtb = DefaultTypedLoggerValue
        except NameError:
            pass
        else:
            if hasattr(cls, '_default_value') and not issubclass(cls, dtb):
                raise TypeError(f'{cls} should be a subclass of DefaultTypedLoggerValue if it has a _default_value')

        dks: ta.Optional[str]
        if isinstance((dk := getattr(cls, '_default_key')), str):
            dks = dk
        elif dk is True:
            dks = snake_case(cls.__name__)
        elif dk is False:
            dks = None
        else:
            raise TypeError(dk)
        cls.__default_key = dks  # type: ignore[attr-defined]

    #

    @ta.final
    def _typed_logger_provide_value(self, ctx: 'TypedLoggerContext') -> TypedLoggerValueOrAbsent:  # noqa
        return self

    @classmethod
    def _typed_logger_maybe_provide_default_value(cls, ctx: 'TypedLoggerContext') -> ta.Tuple[TypedLoggerValueOrAbsent, ...]:  # noqa
        return ()

    @ta.final
    def _typed_logger_resolve_field_value(self, ctx: 'TypedLoggerContext') -> ResolvedTypedLoggerFieldValue:  # noqa
        return self

    @ta.final
    def _typed_logger_unwrap_field_value(self, ctx: 'TypedLoggerContext') -> UnwrappedTypedLoggerFieldValue:  # noqa
        return self._v

    def _typed_logger_visit_bindings(self, vst: 'TypedLoggerBindings._Visitor') -> None:  # noqa
        vst.accept_values(((type(self), self),))
        vst.accept_const_values(((type(self), self),))


class DefaultTypedLoggerValue(TypedLoggerValue[T], Abstract):
    _default_value: ta.ClassVar[ta.Union[
        AbsentTypedLoggerValue,
        classmethod,
        TypedLoggerValue.ContextLambda,
        ta.Any,
    ]] = ABSENT_TYPED_LOGGER_VALUE

    @ta.final
    @classmethod
    def default_provider(cls) -> 'TypedLoggerValueProvider':
        try:
            return cls.__default_provider  # type: ignore[attr-defined]
        except AttributeError:
            pass

        # Must be done late to support typing forwardrefs.
        dp: TypedLoggerValueProvider
        dv = next(mc.__dict__['_default_value'] for mc in cls.__mro__ if '_default_value' in mc.__dict__)

        if dv is ABSENT_TYPED_LOGGER_VALUE:
            dp = ConstTypedLoggerValueProvider(cls, ABSENT_TYPED_LOGGER_VALUE)

        elif isinstance(dv, classmethod):
            fn = dv.__get__(None, cls)
            fl: ta.Any = lambda **kw: cls(v) if (v := fn(**kw)) is not ABSENT_TYPED_LOGGER_VALUE else v
            dp = FnTypedLoggerValueProvider(cls, update_wrapper_no_annotations(fl, fn))

        elif isinstance(dv, TypedLoggerValue.ContextLambda):
            fl = lambda ctx: cls(dv.fn(ctx))
            dp = ContextFnTypedLoggerValueProvider(cls, update_wrapper_no_annotations(fl, dv.fn))

        else:
            dp = ConstTypedLoggerValueProvider(cls, cls(dv))

        cls.__default_provider = dp  # type: ignore[attr-defined]
        return dp

    #

    @ta.final
    @classmethod
    def _typed_logger_maybe_provide_default_value(cls, ctx: 'TypedLoggerContext') -> ta.Tuple[TypedLoggerValueOrAbsent, ...]:  # noqa
        return (cls.default_provider().provide_value(ctx),)


class MultiTypedLoggerValue(TypedLoggerValue[T], Abstract):
    """
    Note: There is no DefaultMultiTypedLoggerValue.
    """

    @classmethod
    @abc.abstractmethod
    def merge_values(cls, *values: T) -> T:
        raise NotImplementedError

    #

    def _typed_logger_visit_bindings(self, vst: 'TypedLoggerBindings._Visitor') -> None:  # noqa
        vst.accept_multi_values(((type(self), self),))


##


class TypedLoggerValueProvider(Abstract, ta.Generic[TypedLoggerValueT]):
    @property
    @abc.abstractmethod
    def cls(self) -> ta.Type[TypedLoggerValueT]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide_value(self, ctx: 'TypedLoggerContext') -> ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]:
        raise NotImplementedError

    #

    @ta.final
    def _typed_logger_provide_value(self, ctx: 'TypedLoggerContext') -> TypedLoggerValueOrAbsent:  # noqa
        return self.provide_value(ctx)

    @ta.final
    def _typed_logger_resolve_field_value(self, ctx: 'TypedLoggerContext') -> ResolvedTypedLoggerFieldValue:  # noqa
        return self.provide_value(ctx)

    def _typed_logger_visit_bindings(self, vst: 'TypedLoggerBindings._Visitor') -> None:  # noqa
        vst.accept_values(((self.cls, self),))


#


class ConstTypedLoggerValueProvider(TypedLoggerValueProvider[TypedLoggerValueT]):
    def __init__(
            self,
            cls: ta.Type[TypedLoggerValueT],
            v: ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue],
    ) -> None:
        super().__init__()

        self._cls = cls
        self._v = v

    @property
    def cls(self) -> ta.Type[TypedLoggerValueT]:
        return self._cls

    @property
    def v(self) -> ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]:
        return self._v

    def provide_value(self, ctx: 'TypedLoggerContext') -> ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]:
        return self._v

    #

    @ta.final
    def _typed_logger_visit_bindings(self, vst: 'TypedLoggerBindings._Visitor') -> None:  # noqa
        vst.accept_values(((self.cls, self),))
        vst.accept_const_values(((self.cls, self._v),))


#


@ta.final
class ContextFnTypedLoggerValueProvider(TypedLoggerValueProvider[TypedLoggerValueT]):
    def __init__(
            self,
            cls: ta.Type[TypedLoggerValueT],
            fn: ta.Callable[['TypedLoggerContext'], ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]],
    ) -> None:
        super().__init__()

        self._cls = cls
        self._fn = fn

    @property
    def cls(self) -> ta.Type[TypedLoggerValueT]:
        return self._cls

    @property
    def fn(self) -> ta.Callable[['TypedLoggerContext'], ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]]:
        return self._fn

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._cls!r}, {self._fn!r})'

    def provide_value(self, ctx: 'TypedLoggerContext') -> ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]:
        return self._fn(ctx)


#


class FnTypedLoggerValueProviderAnnotationError(TypeError):
    pass


@ta.final
class FnTypedLoggerValueProvider(TypedLoggerValueProvider[TypedLoggerValueT]):
    def __init__(
            self,
            cls: ta.Type[TypedLoggerValueT],
            fn: ta.Callable[[], ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]],
    ) -> None:
        super().__init__()

        self._cls = cls
        self._fn = fn

        self._kw = self._get_fn_kwargs(fn)

    @classmethod
    def _get_fn_kwargs(cls, fn: ta.Any) -> ta.Mapping[str, ta.Type[TypedLoggerValue]]:
        o = fn
        s = {o}
        while hasattr(o, '__wrapped__'):
            o = o.__wrapped__
            if o in s:
                raise RuntimeError(f'Recursive unwrap: {o!r} in {s!r}')
            s.add(o)

        def get_kw(anns: ta.Dict[str, ta.Any]) -> ta.Dict[str, ta.Type[TypedLoggerValue]]:
            anns = dict(anns)
            anns.pop('return', None)
            if not anns:
                return {}

            kw: ta.Dict[str, ta.Type[TypedLoggerValue]] = {}
            for k, v in anns.items():
                def bad() -> ta.NoReturn:
                    raise FnTypedLoggerValueProviderAnnotationError(
                        f'{fn} has invalid annotation {k} of {v} - must be subtype of TypedLoggerValue or'  # noqa
                        f'Union[TypedLoggerValue, AbsentTypedLoggerValue]',
                    )

                a = v
                if is_union_alias(a):
                    if len(us := set(a)) != 2 or AbsentTypedLoggerValue not in us:
                        bad()
                    [a] = us - {AbsentTypedLoggerValue}

                if not (isinstance(a, type) and issubclass(a, TypedLoggerValue)):
                    bad()

                kw[k] = a

            return kw

        try:
            # Note: transparently falls back to magic property on 3.14+
            return get_kw(getattr(o, '__annotations__', {}))
        except FnTypedLoggerValueProviderAnnotationError:
            pass

        # This is much slower, so only do it if necessary.
        return get_kw(ta.get_type_hints(o))

    @property
    def cls(self) -> ta.Type[TypedLoggerValueT]:
        return self._cls

    @property
    def fn(self) -> ta.Callable[[], ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]]:
        return self._fn

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._cls!r}, {self._fn!r})'

    def provide_value(self, ctx: 'TypedLoggerContext') -> ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]:
        return self._fn(**{
            k: v
            for k, c in self._kw.items()
            if (v := ctx[c]) is not ABSENT_TYPED_LOGGER_VALUE
        })


##


@ta.final
class TypedLoggerConstFieldValue(ta.NamedTuple):
    v: ta.Any

    #

    @ta.final
    def _typed_logger_resolve_field_value(self, ctx: 'TypedLoggerContext') -> ResolvedTypedLoggerFieldValue:  # noqa
        return self

    @ta.final
    def _typed_logger_unwrap_field_value(self, ctx: 'TypedLoggerContext') -> UnwrappedTypedLoggerFieldValue:  # noqa
        return self.v


@ta.final
class TypedLoggerField(ta.NamedTuple):
    k: str
    v: TypedLoggerFieldValue

    #

    def _typed_logger_visit_bindings(self, vst: 'TypedLoggerBindings._Visitor') -> None:  # noqa
        vst.accept_keys(((self.k, self.v),))


##


TYPED_LOGGER_VALUE_OR_PROVIDER_TYPES: ta.Tuple[ta.Type[TypedLoggerValueOrProvider], ...] = (
    TypedLoggerValue,
    TypedLoggerValueProvider,
)

TYPED_LOGGER_VALUE_OR_ABSENT_TYPES: ta.Tuple[ta.Union[
    ta.Type[TypedLoggerValue],
    AbsentTypedLoggerValue,
], ...] = (
    TypedLoggerValue,
    ABSENT_TYPED_LOGGER_VALUE,
)

TYPED_LOGGER_VALUE_OR_PROVIDER_OR_ABSENT_TYPES: ta.Tuple[ta.Union[
    ta.Type[TypedLoggerValueOrProvider],
    AbsentTypedLoggerValue,
], ...] = (
    *TYPED_LOGGER_VALUE_OR_PROVIDER_TYPES,
    ABSENT_TYPED_LOGGER_VALUE,
)


##


@ta.overload
def unwrap_typed_logger_field_value(rfv: TypedLoggerValue[T]) -> TypedLoggerValue[T]:
    ...


@ta.overload
def unwrap_typed_logger_field_value(rfv: AbsentTypedLoggerValue) -> AbsentTypedLoggerValue:
    ...


@ta.overload
def unwrap_typed_logger_field_value(rfv: TypedLoggerConstFieldValue) -> ta.Any:
    ...


def unwrap_typed_logger_field_value(rfv):
    return rfv._typed_logger_unwrap_field_value(rfv)  # noqa



"""
An abstraction over greenlet's api. Greenlet doesn't currently support nogil but its functionality is needed for async
bridge code (both here and in sqlalchemy). This can be implemented with real threads at the expense of overhead, but
this code is only intended to be used in already fairly heavy situations (bootstrap, db calls).
"""
import abc
import dataclasses as dc
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    import greenlet
else:
    greenlet = lang.proxy_import('greenlet')


##


class Threadlet(lang.Abstract):
    """Not safe to identity-key - use `underlying`."""

    def __hash__(self):
        raise TypeError('use `underlying`')

    def __eq__(self, other):
        raise TypeError('use `underlying`')

    @property
    @abc.abstractmethod
    def underlying(self) -> ta.Any:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def parent(self) -> ta.Optional['Threadlet']:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dead(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def switch(self, *args: ta.Any, **kwargs: ta.Any) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def throw(self, ex: BaseException) -> ta.Any:
        raise NotImplementedError


class Threadlets(lang.Abstract):
    @abc.abstractmethod
    def spawn(self, fn: ta.Callable[[], None]) -> Threadlet:
        raise NotImplementedError

    @abc.abstractmethod
    def get_current(self) -> Threadlet:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class GreenletThreadlet(Threadlet):
    g: 'greenlet.greenlet'

    @property
    def underlying(self) -> 'greenlet.greenlet':
        return self.g

    @property
    def parent(self) -> ta.Optional['GreenletThreadlet']:
        return GreenletThreadlet(p) if (p := self.g.parent) is not None else None

    @property
    def dead(self) -> bool:
        return self.g.dead

    def switch(self, *args: ta.Any, **kwargs: ta.Any) -> ta.Any:
        return self.g.switch(*args, **kwargs)

    def throw(self, ex: BaseException) -> ta.Any:
        return self.g.throw(ex)


class GreenletThreadlets(Threadlets):
    def spawn(self, fn: ta.Callable[[], None]) -> Threadlet:
        return GreenletThreadlet(greenlet.greenlet(fn))

    def get_current(self) -> Threadlet:
        return GreenletThreadlet(greenlet.getcurrent())


GREENLET_THREADLETS = GreenletThreadlets()



# ruff: noqa: UP006 UP045
import functools
import typing as ta


T = ta.TypeVar('T')


##


async def opt_await(aw: ta.Optional[ta.Awaitable[T]]) -> ta.Optional[T]:
    return (await aw if aw is not None else None)


async def async_list(ai: ta.AsyncIterable[T]) -> ta.List[T]:
    return [v async for v in ai]


async def async_enumerate(ai: ta.AsyncIterable[T]) -> ta.AsyncIterable[ta.Tuple[int, T]]:
    i = 0
    async for e in ai:
        yield (i, e)
        i += 1


##


def as_async(fn: ta.Callable[..., T], *, wrap: bool = False) -> ta.Callable[..., ta.Awaitable[T]]:
    async def inner(*args, **kwargs):
        return fn(*args, **kwargs)

    return functools.wraps(fn)(inner) if wrap else inner


##


class SyncAwaitCoroutineNotTerminatedError(Exception):
    pass


def sync_await(aw: ta.Awaitable[T]) -> T:
    """
    Allows for the synchronous execution of async functions which will never actually *externally* await anything. These
    functions are allowed to await any number of other functions - including contextmanagers and generators - so long as
    nothing ever actually 'leaks' out of the function, presumably to an event loop.
    """

    ret = missing = object()

    async def thunk():
        nonlocal ret

        ret = await aw

    cr = thunk()
    try:
        try:
            cr.send(None)
        except StopIteration:
            pass

        if ret is missing or cr.cr_await is not None or cr.cr_running:
            raise SyncAwaitCoroutineNotTerminatedError('Not terminated')

    finally:
        cr.close()

    return ta.cast(T, ret)


#


def sync_aiter(ai: ta.AsyncIterator[T]) -> ta.Iterator[T]:
    while True:
        try:
            o = sync_await(ai.__anext__())
        except StopAsyncIteration:
            break
        yield o


def sync_async_list(ai: ta.AsyncIterable[T]) -> ta.List[T]:
    """
    Uses `sync_await` to synchronously read the full contents of a function call returning an async iterator, given that
    the function never externally awaits anything.
    """

    lst: ta.Optional[ta.List[T]] = None

    async def inner():
        nonlocal lst

        lst = [v async for v in ai]

    sync_await(inner())

    if not isinstance(lst, list):
        raise TypeError(lst)

    return lst


#


@ta.final
class SyncAwaitContextManager(ta.Generic[T]):
    def __init__(self, acm: ta.AsyncContextManager[T]) -> None:
        self._acm = acm

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._acm!r})'

    def __enter__(self) -> T:
        return sync_await(self._acm.__aenter__())

    def __exit__(self, exc_type, exc_val, exc_tb):
        return sync_await(self._acm.__aexit__(exc_type, exc_val, exc_tb))


sync_async_with = SyncAwaitContextManager


##


@ta.final
class SyncToAsyncContextManager(ta.Generic[T]):
    def __init__(self, cm: ta.ContextManager[T]) -> None:
        self._cm = cm

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._cm!r})'

    async def __aenter__(self) -> T:
        return self._cm.__enter__()

    async def __aexit__(self, exc_type, exc_value, traceback, /):
        return self._cm.__exit__(exc_type, exc_value, traceback)


as_async_context_manager = SyncToAsyncContextManager



import typing as ta

from .. import cached
from .. import check
from .. import dataclasses as dc
from .. import lang
from .base import AsyncLifecycle
from .base import Lifecycle
from .states import LifecycleState
from .states import LifecycleStates


LifecycleManagedT = ta.TypeVar('LifecycleManagedT', bound='LifecycleManaged')

AsyncLifecycleManagedT = ta.TypeVar('AsyncLifecycleManagedT', bound='AsyncLifecycleManaged')


##


class LifecycleManaged(lang.Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        try:
            async_lifecycle_managed_cls = AsyncLifecycleManaged
        except NameError:
            pass
        else:
            check.not_issubclass(cls, async_lifecycle_managed_cls)

    @ta.final
    @dc.dataclass(frozen=True)
    class _Lifecycle(
        Lifecycle,
        lang.Final,
        ta.Generic[LifecycleManagedT],
    ):
        obj: LifecycleManagedT

        def lifecycle_state(self, state: LifecycleState) -> None:
            self.obj._lifecycle_state_ = state

        def lifecycle_construct(self) -> None:
            self.obj._lifecycle_construct()  # noqa

        def lifecycle_start(self) -> None:
            self.obj._lifecycle_start()  # noqa

        def lifecycle_stop(self) -> None:
            self.obj._lifecycle_stop()  # noqa

        def lifecycle_destroy(self) -> None:
            self.obj._lifecycle_destroy()  # noqa

    @ta.final
    @cached.property
    def _lifecycle(self) -> _Lifecycle[ta.Self]:
        return LifecycleManaged._Lifecycle(self)

    _lifecycle_state_: LifecycleState = LifecycleStates.NEW

    @property
    def _lifecycle_state(self) -> LifecycleState:
        return self._lifecycle_state_

    def _lifecycle_construct(self) -> None:
        pass

    def _lifecycle_start(self) -> None:
        pass

    def _lifecycle_stop(self) -> None:
        pass

    def _lifecycle_destroy(self) -> None:
        pass


#


class AsyncLifecycleManaged(lang.Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        check.not_issubclass(cls, LifecycleManaged)

    @ta.final
    @dc.dataclass(frozen=True)
    class _Lifecycle(
        AsyncLifecycle,
        lang.Final,
        ta.Generic[AsyncLifecycleManagedT],
    ):
        obj: AsyncLifecycleManagedT

        async def lifecycle_state(self, state: LifecycleState) -> None:
            self.obj._lifecycle_state_ = state

        async def lifecycle_construct(self) -> None:
            await self.obj._lifecycle_construct()  # noqa

        async def lifecycle_start(self) -> None:
            await self.obj._lifecycle_start()  # noqa

        async def lifecycle_stop(self) -> None:
            await self.obj._lifecycle_stop()  # noqa

        async def lifecycle_destroy(self) -> None:
            await self.obj._lifecycle_destroy()  # noqa

    @ta.final
    @cached.property
    def _lifecycle(self) -> _Lifecycle[ta.Self]:
        return AsyncLifecycleManaged._Lifecycle(self)

    _lifecycle_state_: LifecycleState = LifecycleStates.NEW

    @property
    def _lifecycle_state(self) -> LifecycleState:
        return self._lifecycle_state_

    async def _lifecycle_construct(self) -> None:
        pass

    async def _lifecycle_start(self) -> None:
        pass

    async def _lifecycle_stop(self) -> None:
        pass

    async def _lifecycle_destroy(self) -> None:
        pass


#


AnyLifecycleManaged: ta.TypeAlias = LifecycleManaged | AsyncLifecycleManaged

ANY_LIFECYCLE_MANAGED_TYPES: tuple[type[LifecycleManaged | AsyncLifecycleManaged], ...] = (LifecycleManaged, AsyncLifecycleManaged)  # noqa



def cdiv(x: int, y: int) -> int:
    """https://python-history.blogspot.com/2010/08/why-pythons-integer-division-floors.html"""

    if y == 0:
        return 0
    u = abs(x) // abs(y)
    if (x < 0) ^ (y < 0):
        u *= -1
    return u


def cmod(x: int, y: int) -> int:
    return x - (cdiv(x, y) * y)



class Error(Exception):
    pass


##


class ColumnError(Error):
    pass


class DuplicateColumnNameError(ColumnError):
    pass


class MismatchedColumnCountError(ColumnError):
    pass


##


class QueryError(Error):
    pass



import abc
import dataclasses as dc
import itertools
import typing as ta

from .. import lang


T = ta.TypeVar('T')
U = ta.TypeVar('U')

Transform: ta.TypeAlias = ta.Callable[[ta.Iterable[T]], ta.Iterable[U]]

X0 = ta.TypeVar('X0')
X1 = ta.TypeVar('X1')
X2 = ta.TypeVar('X2')
X3 = ta.TypeVar('X3')
X4 = ta.TypeVar('X4')
X5 = ta.TypeVar('X5')


##


_filter = filter
_map = map
_zip = zip


##


class Transform_(lang.Abstract, ta.Generic[T, U]):  # noqa
    @abc.abstractmethod
    def __call__(self, it: ta.Iterable[T]) -> ta.Iterable[U]:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class Filter(Transform_[T, T]):
    fn: ta.Callable[[T], bool] | None

    def __call__(self, it: ta.Iterable[T]) -> ta.Iterable[T]:
        return _filter(self.fn, it)


filter = Filter  # noqa


##


@dc.dataclass(frozen=True)
class Map(Transform_[T, U]):
    fn: ta.Callable[[T], U]

    def __call__(self, it: ta.Iterable[T]) -> ta.Iterable[U]:
        return _map(self.fn, it)


map = Map  # noqa


##


@dc.dataclass(frozen=True)
class Apply(Transform_[T, T]):
    fn: ta.Callable[[T], ta.Any]

    def __call__(self, it: ta.Iterable[T]) -> ta.Iterable[T]:
        return _map(lambda e: (self.fn(e), e)[1], it)


apply = Apply


##


@dc.dataclass(frozen=True)
class Flatten(Transform_[ta.Iterable[T], T]):
    def __call__(self, it: ta.Iterable[ta.Iterable[T]]) -> ta.Iterable[T]:
        return itertools.chain.from_iterable(it)


flatten = Flatten


##


@dc.dataclass(frozen=True)
class FlatMap(Transform_[T, U]):
    fn: ta.Callable[[T], ta.Iterable[U]]

    def __call__(self, it: ta.Iterable[T]) -> ta.Iterable[U]:
        return itertools.chain.from_iterable(_map(self.fn, it))


flat_map = FlatMap

# Alternatively:
# def flat_map(fn: ta.Callable[[T], ta.Iterable[U]]) -> Transform[T, U]:
#     return compose(map(fn), flatten())


##


@dc.dataclass(frozen=True)
class Compose(Transform_[T, U]):
    tfs: ta.Sequence[Transform]

    def __call__(self, it: ta.Iterable[T]) -> ta.Iterable[U]:
        for tf in self.tfs:
            it = tf(it)
        return it  # type: ignore[return-value]


@ta.overload
def compose(
        tf0: Transform[T, U],
) -> Transform[T, U]:
    ...


@ta.overload
def compose(
        tf0: Transform[T, X0],
        tf1: Transform[X0, U],
) -> Transform[T, U]:
    ...


@ta.overload
def compose(
        tf0: Transform[T, X0],
        tf1: Transform[X0, X1],
        tf2: Transform[X1, U],
) -> Transform[T, U]:
    ...


@ta.overload
def compose(
        tf0: Transform[T, X0],
        tf1: Transform[X0, X1],
        tf2: Transform[X1, X2],
        tf3: Transform[X2, U],
) -> Transform[T, U]:
    ...


@ta.overload
def compose(
        tf0: Transform[T, X0],
        tf1: Transform[X0, X1],
        tf2: Transform[X1, X2],
        tf3: Transform[X2, X3],
        tf4: Transform[X3, U],
) -> Transform[T, U]:
    ...


@ta.overload
def compose(
        tf0: Transform[T, X0],
        tf1: Transform[X0, X1],
        tf2: Transform[X1, X2],
        tf3: Transform[X2, X3],
        tf4: Transform[X3, X4],
        tf5: Transform[X4, U],
) -> Transform[T, U]:
    ...


@ta.overload
def compose(
        tf0: Transform[T, X0],
        tf1: Transform[X0, X1],
        tf2: Transform[X1, X2],
        tf3: Transform[X2, X3],
        tf4: Transform[X3, X4],
        tf5: Transform[X4, X5],
        tf6: Transform[X5, U],
) -> Transform[T, U]:
    ...


@ta.overload
def compose(
        *tf: Transform,
) -> Transform:
    ...


def compose(tf0, *tfn):
    if not tfn:
        return tf0
    return Compose([tf0, *tfn])



import dataclasses as dc
import enum
import typing as ta

import pytest

from ..... import check
from ..... import collections as col
from .....docker import all as docker


##


class SwitchState(enum.Enum):
    ENABLED = enum.auto()
    DISABLED = enum.auto()
    ONLY = enum.auto()
    IF_SINGLE = enum.auto()


@dc.dataclass(frozen=True, eq=False)
class Switch:
    name: str
    _default_state: SwitchState | ta.Callable[[pytest.Session], SwitchState]

    _: dc.KW_ONLY

    add_marks: ta.Sequence[ta.Any] | None = None

    def default_state(self, session: pytest.Session) -> SwitchState:
        if isinstance(e := self._default_state, SwitchState):
            return e
        elif callable(e):
            return check.isinstance(e(session), SwitchState)
        else:
            raise TypeError(e)

    @property
    def attr(self) -> str:
        return self.name.replace('-', '_')


##


SWITCHES: ta.Sequence[Switch] = [
    Switch(
        'name',
        lambda _: SwitchState.ENABLED if docker.has_cli() else SwitchState.DISABLED,
    ),

    Switch(
        'docker-guest',
        lambda _: SwitchState.ENABLED if docker.is_likely_in_docker() else SwitchState.DISABLED,
    ),

    Switch(
        'online',
        SwitchState.ENABLED,
    ),

    Switch(
        'integration',
        SwitchState.ENABLED,
    ),

    Switch(
        'high-mem',
        SwitchState.ENABLED,
        add_marks=[
            # https://pytest-xdist.readthedocs.io/en/latest/distribution.html
            pytest.mark.xdist_group('high-mem'),
            pytest.mark.gc_collect_after(),
        ],
    ),

    Switch(
        'slow',
        SwitchState.IF_SINGLE,
    ),
]


##


SWITCHES_BY_NAME: ta.Mapping[str, Switch] = col.make_map_by(lambda sw: sw.name, SWITCHES, strict=True)
SWITCHES_BY_ATTR: ta.Mapping[str, Switch] = col.make_map_by(lambda sw: sw.attr, SWITCHES, strict=True)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
➜  ~ cat /etc/os-release
NAME="Amazon Linux"
VERSION="2"
ID="amzn"
ID_LIKE="centos rhel fedora"
VERSION_ID="2"
PRETTY_NAME="Amazon Linux 2"

➜  ~ cat /etc/os-release
PRETTY_NAME="Ubuntu 22.04.5 LTS"
NAME="Ubuntu"
VERSION_ID="22.04"
VERSION="22.04.5 LTS (Jammy Jellyfish)"
VERSION_CODENAME=jammy
ID=ubuntu
ID_LIKE=debian
UBUNTU_CODENAME=jammy

➜  omlish git:(master) docker run -i python:3.13 cat /etc/os-release
PRETTY_NAME="Debian GNU/Linux 12 (bookworm)"
NAME="Debian GNU/Linux"
VERSION_ID="12"
VERSION="12 (bookworm)"
VERSION_CODENAME=bookworm
ID=debian
"""
import dataclasses as dc
import os.path
import typing as ta


##


@dc.dataclass(frozen=True)
class LinuxOsRelease:
    """
    https://man7.org/linux/man-pages/man5/os-release.5.html
    """

    raw: ta.Mapping[str, str]

    # General information identifying the operating system

    @property
    def name(self) -> str:
        """
        A string identifying the operating system, without a version component, and suitable for presentation to the
        user. If not set, a default of "NAME=Linux" may be used.

        Examples: "NAME=Fedora", "NAME="Debian GNU/Linux"".
        """

        return self.raw['NAME']

    @property
    def id(self) -> str:
        """
        A lower-case string (no spaces or other characters outside of 0-9, a-z, ".", "_" and "-") identifying the
        operating system, excluding any version information and suitable for processing by scripts or usage in generated
        filenames. If not set, a default of "ID=linux" may be used. Note that even though this string may not include
        characters that require shell quoting, quoting may nevertheless be used.

        Examples: "ID=fedora", "ID=debian".
        """

        return self.raw['ID']

    @property
    def id_like(self) -> str:
        """
        A space-separated list of operating system identifiers in the same syntax as the ID= setting. It should list
        identifiers of operating systems that are closely related to the local operating system in regards to packaging
        and programming interfaces, for example listing one or more OS identifiers the local OS is a derivative from. An
        OS should generally only list other OS identifiers it itself is a derivative of, and not any OSes that are
        derived from it, though symmetric relationships are possible. Build scripts and similar should check this
        variable if they need to identify the local operating system and the value of ID= is not recognized. Operating
        systems should be listed in order of how closely the local operating system relates to the listed ones, starting
        with the closest. This field is optional.

        Examples: for an operating system with "ID=centos", an assignment of "ID_LIKE="rhel fedora"" would be
        appropriate. For an operating system with "ID=ubuntu", an assignment of "ID_LIKE=debian" is appropriate.
        """

        return self.raw['ID_LIKE']

    @property
    def pretty_name(self) -> str:
        """
        A pretty operating system name in a format suitable for presentation to the user. May or may not contain a
        release code name or OS version of some kind, as suitable. If not set, a default of "PRETTY_NAME="Linux"" may be
        used

        Example: "PRETTY_NAME="Fedora 17 (Beefy Miracle)"".
        """

        return self.raw['PRETTY_NAME']

    @property
    def cpe_name(self) -> str:
        """
        A CPE name for the operating system, in URI binding syntax, following the Common Platform Enumeration
        Specification[4] as proposed by the NIST. This field is optional.

        Example: "CPE_NAME="cpe:/o:fedoraproject:fedora:17""
        """

        return self.raw['CPE_NAME']

    @property
    def variant(self) -> str:
        """
        A string identifying a specific variant or edition of the operating system suitable for presentation to the
        user. This field may be used to inform the user that the configuration of this system is subject to a specific
        divergent set of rules or default configuration settings. This field is optional and may not be implemented on
        all systems.

        Examples: "VARIANT="Server Edition"", "VARIANT="Smart Refrigerator Edition"".

        Note: this field is for display purposes only. The VARIANT_ID field should be used for making programmatic
        decisions.

        Added in version 220.
        """

        return self.raw['VARIANT']

    @property
    def variant_id(self) -> str:
        """
        A lower-case string (no spaces or other characters outside of 0-9, a-z, ".", "_" and "-"), identifying a
        specific variant or edition of the operating system. This may be interpreted by other packages in order to
        determine a divergent default configuration. This field is optional and may not be implemented on all systems.

        Examples: "VARIANT_ID=server", "VARIANT_ID=embedded".

        Added in version 220.
        """

        return self.raw['variant_id']

    # Information about the version of the operating system

    @property
    def version(self) -> str:
        """
        A string identifying the operating system version, excluding any OS name information, possibly including a
        release code name, and suitable for presentation to the user. This field is optional.

        Examples: "VERSION=17", "VERSION="17 (Beefy Miracle)"".
        """

        return self.raw['VERSION']

    @property
    def version_id(self) -> str:
        """
        A lower-case string (mostly numeric, no spaces or other characters outside of 0-9, a-z, ".", "_" and "-")
        identifying the operating system version, excluding any OS name information or release code name, and suitable
        for processing by scripts or usage in generated filenames. This field is optional.

        Examples: "VERSION_ID=17", "VERSION_ID=11.04".
        """

        return self.raw['VERSION_ID']

    @property
    def version_codename(self) -> str:
        """
        A lower-case string (no spaces or other characters outside of 0-9, a-z, ".", "_" and "-") identifying the
        operating system release code name, excluding any OS name information or release version, and suitable for
        processing by scripts or usage in generated filenames. This field is optional and may not be implemented on all
        systems.

        Examples: "VERSION_CODENAME=buster", "VERSION_CODENAME=xenial".

        Added in version 231.
        """

        return self.raw['VERSION_CODENAME']

    @property
    def build_id(self) -> str:
        """
        A string uniquely identifying the system image originally used as the installation base. In most cases,
        VERSION_ID or IMAGE_ID+IMAGE_VERSION are updated when the entire system image is replaced during an update.
        BUILD_ID may be used in distributions where the original installation image version is important: VERSION_ID
        would change during incremental system updates, but BUILD_ID would not. This field is optional.

        Examples: "BUILD_ID="2013-03-20.3"", "BUILD_ID=201303203".

        Added in version 200.
        """

        return self.raw['BUILD_ID']

    @property
    def image_id(self) -> str:
        """
        A lower-case string (no spaces or other characters outside of 0-9, a-z, ".", "_" and "-"), identifying a
        specific image of the operating system. This is supposed to be used for environments where OS images are
        prepared, built, shipped and updated as comprehensive, consistent OS images. This field is optional and may not
        be implemented on all systems, in particularly not on those that are not managed via images but put together and
        updated from individual packages and on the local system.

        Examples: "IMAGE_ID=vendorx-cashier-system", "IMAGE_ID=netbook-image".

        Added in version 249.
        """

        return self.raw['IMAGE_ID']

    @property
    def image_version(self) -> str:
        """
        A lower-case string (mostly numeric, no spaces or other characters outside of 0-9, a-z, ".", "_" and "-")
        identifying the OS image version. This is supposed to be used together with IMAGE_ID described above, to discern
        different versions of the same image.

        Examples: "IMAGE_VERSION=33", "IMAGE_VERSION=47.1rc1".

        Added in version 249.
        """

        return self.raw['IMAGE_VERSION']

    # To summarize: if the image updates are built and shipped as comprehensive units, IMAGE_ID+IMAGE_VERSION is the
    # best fit. Otherwise, if updates eventually completely replace previously installed contents, as in a typical
    # binary distribution, VERSION_ID should be used to identify major releases of the operating system.  BUILD_ID may
    # be used instead or in addition to VERSION_ID when the original system image version is important.

    #

    # Presentation information and links

    # Links to resources on the Internet related to the operating system.  HOME_URL= should refer to the homepage of the
    # operating system, or alternatively some homepage of the specific version of the operating system.
    # DOCUMENTATION_URL= should refer to the main documentation page for this operating system.  SUPPORT_URL= should
    # refer to the main support page for the operating system, if there is any. This is primarily intended for operating
    # systems which vendors provide support for.  BUG_REPORT_URL= should refer to the main bug reporting page for the
    # operating system, if there is any. This is primarily intended for operating systems that rely on community QA.
    # PRIVACY_POLICY_URL= should refer to the main privacy policy page for the operating system, if there is any. These
    # settings are optional, and providing only some of these settings is common. These URLs are intended to be exposed
    # in "About this system" UIs behind links with captions such as "About this Operating System", "Obtain Support",
    # "Report a Bug", or "Privacy Policy". The values should be in RFC3986 format[5], and should be "http:" or "https:"
    # URLs, and possibly "mailto:" or "tel:". Only one URL shall be listed in each setting. If multiple resources need
    # to be referenced, it is recommended to provide an online landing page linking all available resources.

    # Examples: "HOME_URL="https://fedoraproject.org/"", "BUG_REPORT_URL="https://bugzilla.redhat.com/"".

    @property
    def home_url(self) -> str:
        return self.raw['HOME_URL']

    @property
    def documentation_url(self) -> str:
        return self.raw['DOCUMENTATION_URL']

    @property
    def support_url(self) -> str:
        return self.raw['SUPPORT_URL']

    @property
    def bug_report_url(self) -> str:
        return self.raw['BUG_REPORT_URL']

    @property
    def privacy_policy_url(self) -> str:
        return self.raw['PRIVACY_POLICY_URL']

    @property
    def support_end(self) -> str:
        """
        The date at which support for this version of the OS ends. (What exactly "lack of support" means varies between
        vendors, but generally users should assume that updates, including security fixes, will not be provided.) The
        value is a date in the ISO 8601 format "YYYY-MM-DD", and specifies the first day on which support is not
        provided.

        For example, "SUPPORT_END=2001-01-01" means that the system was supported until the end of the last day of the
        previous millennium.

        Added in version 252.
        """

        return self.raw['SUPPORT_END']

    @property
    def logo(self) -> str:
        """
        A string, specifying the name of an icon as defined by freedesktop.org Icon Theme Specification[6]. This can be
        used by graphical applications to display an operating system's or distributor's logo. This field is optional
        and may not necessarily be implemented on all systems.

        Examples: "LOGO=fedora-logo", "LOGO=distributor-logo-opensuse"

        Added in version 240.
        """

        return self.raw['LOGO']

    @property
    def ansi_color(self) -> str:
        """
        A suggested presentation color when showing the OS name on the console. This should be specified as string
        suitable for inclusion in the ESC [ m ANSI/ECMA-48 escape code for setting graphical rendition. This field is
        optional.

        Examples: "ANSI_COLOR="0;31"" for red, "ANSI_COLOR="1;34"" for light blue, or "ANSI_COLOR="0;38;2;60;110;180""
        for Fedora blue.
        """

        return self.raw['ANSI_COLOR']

    @property
    def vendor_name(self) -> str:
        """
        The name of the OS vendor. This is the name of the organization or company which produces the OS. This field is
        optional.

        This name is intended to be exposed in "About this system" UIs or software update UIs when needed to distinguish
        the OS vendor from the OS itself. It is intended to be human readable.

        Examples: "VENDOR_NAME="Fedora Project"" for Fedora Linux, "VENDOR_NAME="Canonical"" for Ubuntu.

        Added in version 254.
        """

        return self.raw['VENDOR_NAME']

    @property
    def vendor_url(self) -> str:
        """
        The homepage of the OS vendor. This field is optional. The VENDOR_NAME= field should be set if this one is,
        although clients must be robust against either field not being set.

        The value should be in RFC3986 format[5], and should be "http:" or "https:" URLs. Only one URL shall be listed
        in the setting.

        Examples: "VENDOR_URL="https://fedoraproject.org/"", "VENDOR_URL="https://canonical.com/"".

        Added in version 254.
        """

        return self.raw['VENDOR_URL']

    # Distribution-level defaults and metadata

    @property
    def default_hostname(self) -> str:
        """
        A string specifying the hostname if hostname(5) is not present and no other configuration source specifies the
        hostname. Must be either a single DNS label (a string composed of 7-bit ASCII lower-case characters and no
        spaces or dots, limited to the format allowed for DNS domain name labels), or a sequence of such labels
        separated by single dots that forms a valid DNS FQDN. The hostname must be at most 64 characters, which is a
        Linux limitation (DNS allows longer names).

        See org.freedesktop.hostname1(5) for a description of how systemd-hostnamed.service(8) determines the fallback
        hostname.

        Added in version 248.
        """

        return self.raw['DEFAULT_HOSTNAME']

    @property
    def architecture(self) -> str:
        """
        A string that specifies which CPU architecture the userspace binaries require. The architecture identifiers are
        the same as for ConditionArchitecture= described in systemd.unit(5). The field is optional and should only be
        used when just single architecture is supported. It may provide redundant information when used in a GPT
        partition with a GUID type that already encodes the architecture. If this is not the case, the architecture
        should be specified in e.g., an extension image, to prevent an incompatible host from loading it.

        Added in version 252.
        """

        return self.raw['ARCHITECTURE']

    @property
    def sysext_level(self) -> str:
        """
        A lower-case string (mostly numeric, no spaces or other characters outside of 0-9, a-z, ".", "_" and "-")
        identifying the operating system extensions support level, to indicate which extension images are supported. See
        /usr/lib/extension-release.d/extension-release.IMAGE, initrd[2] and systemd-sysext(8)) for more information.

        Examples: "SYSEXT_LEVEL=2", "SYSEXT_LEVEL=15.14".

        Added in version 248.
        """

        return self.raw['SYSEXT_LEVEL']

    @property
    def confext_level(self) -> str:
        """
        Semantically the same as SYSEXT_LEVEL= but for confext images. See
        /etc/extension-release.d/extension-release.IMAGE for more information.

        Examples: "CONFEXT_LEVEL=2", "CONFEXT_LEVEL=15.14".

        Added in version 254.
        """

        return self.raw['CONFEXT_LEVEL']

    @property
    def sysext_scope(self) -> str:
        """
        Takes a space-separated list of one or more of the strings "system", "initrd" and "portable". This field is only
        supported in extension-release.d/ files and indicates what environments the system extension is applicable to:
        i.e. to regular systems, to initrds, or to portable service images. If unspecified, "SYSEXT_SCOPE=system
        portable" is implied, i.e. any system extension without this field is applicable to regular systems and to
        portable service environments, but not to initrd environments.

        Added in version 250.
        """

        return self.raw['SYSEXT_SCOPE']

    @property
    def confext_scope(self) -> str:
        """
        Semantically the same as SYSEXT_SCOPE= but for confext images.

        Added in version 254.
        """

        return self.raw['CONFEXT_SCOPE']

    @property
    def portable_prefixes(self) -> str:
        """
        Takes a space-separated list of one or more valid prefix match strings for the Portable Services[3] logic. This
        field serves two purposes: it is informational, identifying portable service images as such (and thus allowing
        them to be distinguished from other OS images, such as bootable system images). It is also used when a portable
        service image is attached: the specified or implied portable service prefix is checked against the list
        specified here, to enforce restrictions how images may be attached to a system.

        Added in version 250.
        """

        return self.raw['PORTABLE_PREFIXES']

    #

    DEFAULT_PATHS: ta.ClassVar[ta.Sequence[str]] = [
        '/etc/os-release',
        '/usr/lib/os-release',
    ]

    @classmethod
    def read(cls, *paths: str) -> ta.Optional['LinuxOsRelease']:
        for fp in (paths or cls.DEFAULT_PATHS):
            if not os.path.isfile(fp):
                continue
            with open(fp) as f:
                src = f.read()
            break
        else:
            return None

        raw = cls._parse_os_release(src)

        return cls(raw)

    @classmethod
    def _parse_os_release(cls, src: str) -> ta.Mapping[str, str]:
        dct: ta.Dict[str, str] = {}

        for l in src.splitlines():
            if not (l := l.strip()):
                continue
            if l.startswith('#') or '=' not in l:
                continue

            k, _, v = l.partition('=')
            if k.startswith('"'):
                k = k[1:-1]
            if v.startswith('"'):
                v = v[1:-1]

            dct[k] = v

        return dct



import typing as ta


Value: ta.TypeAlias = ta.Union[  # noqa
    None,

    bool,
    int,
    float,
    # Number,
    str,
    bytes,

    list,  # list[Value],
    dict,  # dict[str, Value],

    # ta.Any,
]



"""
TODO:
 - parser?
 - js? viz.js, d3, visjs
 - cycle detection
 - networkx adapter
 - https://docs.python.org/3.9/library/graphlib.html#module-graphlib
"""
import typing as ta

from .. import check
from .. import lang


T = ta.TypeVar('T')


##


class LinkError(KeyError):
    pass


def traverse_links(
        links: ta.Mapping[T, ta.Iterable[T]],
        roots: ta.Iterable[T],
        *,
        include_roots: bool = False,
        strict: bool = False,
) -> set[T]:
    """Returns all keys deeply reachable from given roots. Handles cycles."""

    roots = set(roots)

    todo = set(roots)
    seen: set[T] = set()
    while todo:
        key = todo.pop()
        seen.add(key)

        try:
            cur = links[key]
        except KeyError:
            if strict:
                raise LinkError(key) from None
        else:
            todo.update(set(cur) - seen)

    if include_roots:
        return seen
    else:
        return seen - roots


def invert_links(
        links: ta.Mapping[T, ta.Iterable[T]],
        *,
        auto_add_roots: bool = False,
        if_absent: ta.Literal['raise', 'ignore', 'add'] = 'add',
) -> dict[T, set[T]]:
    check.in_(if_absent, ('raise', 'ignore', 'add'))
    if if_absent != 'add':
        check.arg(auto_add_roots, 'auto_add_roots must be True with given if_absent is not "add"')

    ret: dict[T, set[T]]
    if auto_add_roots:
        ret = {src: set() for src in links}
    else:
        ret = {}

    for src, dsts in links.items():
        for dst in dsts:
            try:
                tgt = ret[dst]
            except KeyError:
                if if_absent == 'raise':
                    raise LinkError(dst) from None
                elif if_absent == 'ignore':
                    continue
                elif if_absent == 'add':
                    tgt = ret[dst] = set()
                else:
                    raise RuntimeError from None

            tgt.add(src)

    return ret


##


class Dag(ta.Generic[T]):
    """Given 'input_its_by_outputs', or a map from nodes to that node's dependencies."""

    def __init__(
            self,
            input_its_by_outputs: ta.Mapping[T, ta.Iterable[T]],
            *,
            auto_add_outputs: bool = True,
            if_absent: ta.Literal['raise', 'ignore', 'add'] = 'add',
    ) -> None:
        super().__init__()

        self._input_sets_by_output = {u: set(d) for u, d in input_its_by_outputs.items()}

        self._output_sets_by_input = invert_links(
            self._input_sets_by_output,
            auto_add_roots=auto_add_outputs,
            if_absent=if_absent,
        )

    @property
    def input_sets_by_output(self) -> ta.Mapping[T, ta.AbstractSet[T]]:
        return self._input_sets_by_output

    @property
    def output_sets_by_input(self) -> ta.Mapping[T, ta.AbstractSet[T]]:
        return self._output_sets_by_input

    def subdag(
            self,
            roots: ta.Iterable[T],
            *,
            ignored: ta.Iterable[T] | None = None,
    ) -> 'Subdag[T]':
        return Subdag(
            self,
            roots,
            ignored=ignored,
        )


class Subdag(ta.Generic[T]):
    def __init__(
            self,
            dag: Dag[T],
            roots: ta.Iterable[T],
            *,
            ignored: ta.Iterable[T] | None = None,
    ) -> None:
        super().__init__()

        self._dag: Dag[T] = check.isinstance(dag, Dag)
        self._roots = set(roots)
        self._ignored = set(ignored or []) - self._roots

    @property
    def dag(self) -> Dag[T]:
        return self._dag

    @property
    def roots(self) -> ta.AbstractSet[T]:
        return self._roots

    @property
    def ignored(self) -> ta.AbstractSet[T]:
        return self._ignored

    @lang.cached_property
    def inputs(self) -> ta.AbstractSet[T]:
        return traverse_links(self._dag.input_sets_by_output, self._roots) - self._ignored

    @lang.cached_property
    def outputs(self) -> ta.AbstractSet[T]:
        return traverse_links(self._dag.output_sets_by_input, self._roots) - self._ignored

    @lang.cached_property
    def output_inputs(self) -> ta.AbstractSet[T]:
        return traverse_links(self._dag.input_sets_by_output, self.outputs) - self._ignored

    @lang.cached_property
    def all(self) -> ta.AbstractSet[T]:
        return self.roots | self.inputs | self.outputs | self.output_inputs



# ruff: noqa: UP006 UP007 UP045 UP046
# @omlish-lite
import typing as ta

from ..lite.asyncs import sync_await
from ..lite.check import check
from .base import AsyncLogger
from .base import CaptureLoggingContext
from .base import Logger
from .base import LoggingMsgFn
from .contexts import CaptureLoggingContextImpl
from .infos import LoggingContextInfos
from .levels import LogLevel


##


class AsyncLoggerToLogger(Logger):
    def __init__(self, u: AsyncLogger) -> None:
        super().__init__()

        self._u = u

    def get_effective_level(self) -> LogLevel:
        return self._u.get_effective_level()

    def _log(
            self,
            ctx: CaptureLoggingContext,
            msg: ta.Union[str, tuple, LoggingMsgFn],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> None:
        # Nope out early to avoid sync_await if possible - don't bother in the LoggerToAsyncLogger.
        if not self.is_enabled_for(ctx.must_get_info(LoggingContextInfos.Level).level):
            return

        # Note: we hardcode the stack offset of sync_await (which is 2 - sync_await + sync_await.thunk). In non-lite
        # code, lang.sync_await uses a cext if present to avoid being on the py stack, which would obviously complicate
        # this, but this is lite code so we will always have the non-c version.
        sync_await(
            self._u._log(  # noqa
                check.isinstance(ctx, CaptureLoggingContextImpl).inc_stack_offset(3),
                msg,
                *args,
                **kwargs,
            ),
        )


class LoggerToAsyncLogger(AsyncLogger):
    def __init__(self, u: Logger) -> None:
        super().__init__()

        self._u = u

    def get_effective_level(self) -> LogLevel:
        return self._u.get_effective_level()

    async def _log(
            self,
            ctx: CaptureLoggingContext,
            msg: ta.Union[str, tuple, LoggingMsgFn],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> None:
        return self._u._log(  # noqa
            check.isinstance(ctx, CaptureLoggingContextImpl).inc_stack_offset(),
            msg,
            *args,
            **kwargs,
        )



# ruff: noqa: UP006 UP045
# @omlish-lite
import os
import threading
import typing as ta

from ..lite.marshal import unmarshal_obj
from .loading import ManifestLoader


T = ta.TypeVar('T')


##


class GlobalManifestLoader:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    def __init_subclass__(cls, **kwargs):  # noqa
        raise TypeError

    ##

    _lock: ta.ClassVar[threading.RLock] = threading.RLock()

    _instance: ta.ClassVar[ta.Optional[ManifestLoader]] = None

    @classmethod
    def instance(cls) -> ManifestLoader:
        if (inst := cls._instance) is None:
            with cls._lock:
                if (inst := cls._instance) is None:
                    inst = cls._instance = ManifestLoader(cls.default_config())

        return inst

    @classmethod
    def initialize(cls, config: ta.Optional[ManifestLoader.Config] = None) -> ManifestLoader:
        with cls._lock:
            if cls._instance is not None:
                raise RuntimeError(f'{cls.__name__} already initialized')

            if config is None:
                config = cls.default_config()

            inst = cls._instance = ManifestLoader(config)

        return inst

    ##

    @classmethod
    def default_value_instantiator(cls, obj_cls: type, **kwargs: ta.Any) -> ta.Any:
        return unmarshal_obj(kwargs, obj_cls)

    @classmethod
    def default_config(cls) -> ManifestLoader.Config:
        return ManifestLoader.Config(
            discover_packages=True,
            discover_packages_fallback_scan_root_dirs=[os.getcwd()],

            value_instantiator=cls.default_value_instantiator,
        )

    ##

    @classmethod
    def load(
            cls,
            *,
            packages: ta.Optional[ta.Collection[str]] = None,
            classes: ta.Optional[ta.Collection[type]] = None,
    ) -> ta.Sequence[ManifestLoader.LoadedManifest]:
        return cls.instance().load(
            packages=packages,
            classes=classes,
        )

    @classmethod
    def load_values(
            cls,
            *,
            packages: ta.Optional[ta.Collection[str]] = None,
            classes: ta.Optional[ta.Collection[type]] = None,
    ) -> ta.Sequence[ta.Any]:
        return cls.instance().load_values(
            packages=packages,
            classes=classes,
        )

    @classmethod
    def load_values_of(
            cls,
            clz: ta.Type[T],
            *,
            packages: ta.Optional[ta.Collection[str]] = None,
    ) -> ta.Sequence[T]:
        return cls.instance().load_values_of(
            clz,
            packages=packages,
        )



"""
A minor tidiness for hiding refs to [Async]Injector in circ-dep workarounds. Unfortunately unsuitable outside of
injection modules due to it itself being in here (as user code shouldn't reference injector guts), but at least it's
something.
"""
import abc
import functools
import typing as ta

from ... import lang
from ... import reflect as rfl
from ..binder import bind
from ..elements import Elements
from ..elements import as_elements
from ..injector import AsyncInjector
from ..inspect import KwargsTarget
from ..keys import Key
from ..keys import as_key
from ..sync import Injector


T = ta.TypeVar('T')


##


class Late(lang.Abstract, ta.Generic[T]):
    @abc.abstractmethod
    def __call__(self) -> T:
        raise NotImplementedError


class AsyncLate(lang.Abstract, ta.Generic[T]):
    @abc.abstractmethod
    def __call__(self) -> ta.Awaitable[T]:
        raise NotImplementedError


#


def _bind_late(
        injector_cls: type,
        late_cls: ta.Any,
        inner: ta.Any,
        outer: ta.Any | None = None,
) -> Elements:
    inner_key = as_key(inner)

    outer_key: Key
    outer_fac: ta.Callable[[ta.Any], ta.Any]
    if outer is None:
        inner_ann = rfl.to_annotation(inner_key.ty)
        outer_ann = late_cls[inner_ann]
        outer_key = Key(rfl.type_(outer_ann), tag=inner_key.tag)
        outer_fac = lang.identity
    else:
        outer_key = as_key(outer)
        outer_cls = rfl.get_concrete_type(outer_key.ty)
        outer_fac = outer_cls  # type: ignore[assignment]

    return as_elements(
        bind(outer_key, to_fn=KwargsTarget.of(  # noqa
            lambda i: outer_fac(functools.partial(i.provide, inner_key)),
            i=injector_cls,
        )),
    )


def bind_late(inner: ta.Any, outer: ta.Any | None = None) -> Elements:
    return _bind_late(Injector, Late, inner, outer)


def bind_async_late(inner: ta.Any, outer: ta.Any | None = None) -> Elements:
    return _bind_late(AsyncInjector, AsyncLate, inner, outer)



"""
TODO:
 - a reflowing.TextReflower
 - import textwrap lol
 - obviously handle hyphens, underscores, etc
 - optionally preserve/normalize inter-sentence spaces - '.  ' vs '. '
 - detect if 'intentionally' smaller than current remaining line width, if so do not merge.
  - maybe if only ending with punctuation?
 - detect 'matched pairs' of 'quotes'? to preserve whitespace? `foo  bar` ...
  - how to escape it lol - if we see any \\` do we give up?
"""
import re
import typing as ta


SpanKind: ta.TypeAlias = ta.Literal[
    'word',
    'space',
    'symbol',
]


##


class Span(ta.NamedTuple):
    k: SpanKind
    s: str

    def __repr__(self) -> str:
        return f'{self.k}:{self.s!r}'


_SPAN_PAT = re.compile(r'(\w+)|(\s+)')


def split_line_spans(s: str) -> list[Span]:
    spans: list[Span] = []
    p = 0
    for m in _SPAN_PAT.finditer(s):
        l, r = m.span()
        if p < l:
            spans.append(Span('symbol', s[p:l]))
        if m.group(1):
            spans.append(Span('word', m.group(1)))
        else:
            spans.append(Span('space', m.group(2)))
        p = r
    if p < len(s):
        spans.append(Span('symbol', s[p:]))
    return spans


def _main() -> None:
    print(split_line_spans('hi i am a string! this has a hy-phen.'))


if __name__ == '__main__':
    _main()



import logging
import typing as ta

from ..formatters import LoggingContextFormatter
from .records import LoggingContextLogRecord
from .records import LogRecordLoggingContext


##


@ta.final
class StdLoggingFormatter(logging.Formatter):
    def __init__(self, ctx_formatter: LoggingContextFormatter) -> None:
        super().__init__()

        self._ctx_formatter = ctx_formatter

    def format(self, rec: logging.LogRecord) -> str:
        if isinstance(rec, LoggingContextLogRecord):
            ctx = rec._logging_context  # noqa
        else:
            ctx = LogRecordLoggingContext(rec)

        return self._ctx_formatter.format(ctx)
